"""Basic Extractor that is purely example driven."""

import json
import logging
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, List

from curategpt.extract.extractor import AnnotatedObject, Extractor

logger = logging.getLogger(__name__)


@dataclass
class RecursiveExtractor(Extractor):
    """
    Extractor that recursively extracts objects from text.

    See SPIRES
    """

    serialization_format: str = "json"
    model_name: str = "gpt-3.5-turbo"

    def extract(
        self,
        text: str,
        target_class: str,
        examples: List[AnnotatedObject] = None,
        path=None,
        **kwargs,
    ) -> AnnotatedObject:
        if path is None:
            path = []
        if path:
            pathstr = f"Path: {'/'.join(path)}\n"
        else:
            pathstr = ""
        sv = self.schemaview
        prompt = (
            f"Extract a {target_class} object from text in {self.serialization_format} format,\n"
        )
        prompt += "Conforming to the following schema:\n"
        for slot in sv.class_induced_slots(target_class):
            desc = f"## {slot.description}" if slot.description else ""
            if slot.range:
                if slot.range == "string":
                    rng = '"<text>"'
                else:
                    rng = f"<{slot.range}>"
            else:
                rng = "<VALUE>"
            if slot.multivalued:
                rng = f"[{rng}, ...]"
            prompt += f"{slot.name}: {rng} {desc}\n"
        if examples:
            prompt += "Examples:\n\n"
            for example in examples:
                prompt += f"##\nText: {example.text}\n"
                prompt += pathstr
                prompt += f"Response: {self.partially_serialize(example.object, path)}\n"
        prompt += f"\n##\nText: {text}\n\n"
        prompt += pathstr
        prompt += "Response: "
        model = self.model
        print(f"Prompt: {prompt}")
        response = model.prompt(prompt)
        partial_object = self.deserialize(response.text())
        print(f"PO: {partial_object}")
        for slot in sv.class_induced_slots(target_class):
            if slot.range not in sv.all_classes():
                continue
            v = partial_object.object.get(slot.name, None)
            if v:
                if isinstance(v, list):
                    vs = v
                else:
                    vs = [v]
                sub_objects = [
                    self.extract(
                        x, target_class=slot.range, examples=examples, path=path + [slot.name]
                    )
                    for x in vs
                ]
                if isinstance(v, list):
                    partial_object.object[slot.name] = [
                        sub_object.object for sub_object in sub_objects
                    ]
                else:
                    partial_object.object[slot.name] = sub_objects[0].object
        return AnnotatedObject(object=partial_object.object, annotations={"text": text})

    def partially_serialize(self, object: Any, path: List[str]) -> str:
        if isinstance(object, list):
            return "[" + ", ".join([self.partially_serialize(o, path) for o in object]) + "]"
        if len(path) == 0:
            partial_object = deepcopy(object)
            for k, v in partial_object.items():
                if isinstance(v, list):
                    if len(v) > 0 and isinstance(v[0], dict):
                        partial_object[k] = [self._dict_as_str(o) for o in v]
                elif isinstance(v, dict):
                    partial_object[k] = self._dict_as_str(v)
            return json.dumps(partial_object)
        else:
            nxt = path[0]
            return self.partially_serialize(object[nxt], path[1:])

    def _dict_as_str(self, d: Dict[str, Any]) -> str:
        return ", ".join([f"{k} is {v}" for k, v in d.items()])

    def deserialize(self, text: str) -> AnnotatedObject:
        print(f"Parsing: {text}")
        obj = json.loads(text)
        print(f"Direct object: {obj}")
        ao = AnnotatedObject(object=obj)
        print(f"AO object: {ao.object}")
        return ao



import logging
from collections import defaultdict
from dataclasses import dataclass, field
from functools import lru_cache
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import requests
import requests_cache
import yaml
from oaklib import BasicOntologyInterface, get_adapter
from oaklib.interfaces.association_provider_interface import AssociationProviderInterface

from curategpt.formatters.format_utils import camelify
from curategpt.wrappers import BaseWrapper
from curategpt.wrappers.literature import PubmedWrapper

BASE_URL = "https://go-public.s3.amazonaws.com/files/"
INDEX_URL = "https://go-public.s3.amazonaws.com/files/gocam-models.json"
# MODEL_URL_TEMPLATE = "https://go-public.s3.amazonaws.com/files/go-cam/{model_id}.json"
GOCAM_ENDPOINT = "https://api.geneontology.org/api/go-cam/"

ENABLED_BY = "RO:0002333"
PART_OF = "BFO:0000050"
OCCURS_IN = "BFO:0000066"

logger = logging.getLogger(__name__)


@lru_cache
def _relation_id(p: str) -> str:
    ro_adapter = get_adapter("sqlite:obo:ro")
    lbl = ro_adapter.label(p)
    return camelify(lbl) if lbl else p


def _annotations(obj: Dict) -> Dict:
    def _normalize_property(prop: str) -> str:
        if "/" in prop:
            return prop.split("/")[-1]
        return prop

    return {_normalize_property(a["key"]): a["value"] for a in obj["annotations"]}


MAIN_TYPES = [
    "MolecularFunction",
    "InformationBiomacromolecule",
    "BiologicalProcess",
    "CellularComponent",
    "Evidence",
    "ChemicalEntity",
    "AnatomicalEntity",
]


@dataclass
class GOCAMWrapper(BaseWrapper):
    """
    An view over a GO CAM source.
    """

    name: ClassVar[str] = "gocam"

    default_object_type = "Pathway"

    _label_adapter: BasicOntologyInterface = None

    pubmed_wrapper: PubmedWrapper = None

    ro_adapter: BasicOntologyInterface = None

    session: requests.Session = field(default_factory=lambda: requests.Session())

    include_standard_annotations: bool = False

    def __post_init__(self):
        self.pubmed_wrapper = PubmedWrapper()
        self.pubmed_wrapper.set_cache("gocam_pubmed_cache")
        self.ro_adapter = get_adapter("sqlite:obo:ro")
        self.session = requests_cache.CachedSession("gocam_s3_cache")

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        models = requests.get(INDEX_URL).json()
        for model in models:
            if "gocam" not in model:
                raise ValueError(f"Missing gocam in {model}")
            gocam = model["gocam"]
            yield self.object_by_id(gocam.replace("http://model.geneontology.org/", ""))

    def object_by_id(self, object_id: str) -> Optional[Dict]:
        session = self.session
        if not object_id:
            raise ValueError(f"Missing object ID: {object_id}")
        logger.info(f"Fetch object: {object_id}")
        # response = session.get(MODEL_URL_TEMPLATE.format(model_id=object_id))
        local_id = object_id.replace("gocam:", "")
        response = session.get(f"{GOCAM_ENDPOINT}/{local_id}")
        response.raise_for_status()
        obj = response.json()
        return self.object_from_dict(obj)

    def object_from_dict(self, obj: Dict) -> Optional[Dict]:
        id = obj["id"]
        logger.info(f"Processing model id: {id}")
        logger.debug(f"Model: {yaml.dump(obj)}")
        individuals = obj["individuals"]
        individual_to_type = {}
        individual_to_term = {}
        individual_to_annotations = {}
        sources = set()

        lbl2id = {}

        derived_standard_annotations = []

        def _cls(obj: Dict) -> Optional[str]:
            if obj.get("type", None) == "complement":
                # class expression representing NOT
                return None
            if "id" not in obj:
                raise ValueError(f"No ID for {obj}")
            id = obj["id"]
            pfx = id.split(":")[0]
            if pfx in ["UniProtKB", "MGI", "ZFIN", "SGD"]:
                toks = obj["label"].split(" ")
                n = toks[0]
            else:
                n = camelify(obj["label"])
            lbl2id[n] = str(id)
            return n

        for i in individuals:
            typs = [_cls(x) for x in i.get("root-type", [])]
            typs = [x for x in typs if x]
            typ = None
            for t in typs:
                if t in MAIN_TYPES:
                    typ = t
                    break
            if not typ:
                logger.warning(f"Could not find type for {i}")
                continue
            individual_to_type[i["id"]] = typ
            terms = [_cls(x) for x in i.get("type", [])]
            terms = [x for x in terms if x]
            if len(terms) > 1:
                logger.warning(f"Multiple terms for {i}: {terms}")
            if not terms:
                logger.warning(f"No terms for {i}")
                continue
            individual_to_term[i["id"]] = terms[0]
            anns = _annotations(i)
            individual_to_annotations[i["id"]] = anns
            if "source" in anns:
                sources.add(anns["source"])
        activities = []
        relationships = []
        gene_ids = set()
        process_ids = set()
        activities_by_mf_id = defaultdict(list)
        facts_by_property = defaultdict(list)
        for fact in obj["facts"]:
            facts_by_property[fact["property"]].append(fact)
        for fact in facts_by_property.get(ENABLED_BY, []):
            s, o = fact["subject"], fact["object"]
            anns = _annotations(fact)
            evidence = anns.get("evidence", None)
            if evidence:
                pmid = individual_to_annotations.get(evidence, {}).get("source", None)
            else:
                pmid = None
            if s not in individual_to_term:
                logger.warning(f"Missing subject {s} in {individual_to_term}")
                continue
            if o not in individual_to_term:
                logger.warning(f"Missing object {o} in {individual_to_term}")
                continue
            gene_id = individual_to_term[o]
            # we only include a single pmid for each activity
            activity = {
                "gene": gene_id,
                "activity": individual_to_term[s],
                "reference": pmid,
            }
            gene_ids.add(gene_id)
            activities.append(activity)
            activities_by_mf_id[s].append(activity)

        for fact in facts_by_property.get(PART_OF, []):
            s, o = fact["subject"], fact["object"]
            if o not in individual_to_term:
                logger.warning(f"Missing {o} in {individual_to_term}")
                continue
            for a in activities_by_mf_id.get(s, []):
                a["process"] = individual_to_term[o]
                process_ids.add(individual_to_term[o])

        for fact in facts_by_property.get(OCCURS_IN, []):
            s, o = fact["subject"], fact["object"]
            if o not in individual_to_term:
                logger.warning(f"Missing {o} in {individual_to_term}")
                continue
            for a in activities_by_mf_id.get(s, []):
                a["location"] = individual_to_term[o]

        for p, facts in facts_by_property.items():
            for fact in facts:
                s, o = fact["subject"], fact["object"]
                sas = activities_by_mf_id.get(s, [])
                oas = activities_by_mf_id.get(o, [])
                if not sas or not oas:
                    continue
                if individual_to_type.get(s, None) != "MolecularFunction":
                    continue
                if individual_to_type.get(o, None) != "MolecularFunction":
                    continue
                sa = sas[0]
                oa = oas[0]
                if "relationships" not in sa:
                    sa["relationships"] = []
                rel = {
                    "type": _relation_id(p),
                    "target_gene": oa["gene"],
                    "target_activity": oa["activity"],
                }
                if rel not in sa["relationships"]:
                    sa["relationships"].append(rel)
                relationships.append(rel)
        pmids = {a["reference"] for a in activities if "reference" in a}
        pmids = [p for p in pmids if p and p.startswith("PMID")]
        pubs = self.pubmed_wrapper.objects_by_ids(pmids)
        pubs_by_id = {p["id"]: p for p in pubs}
        for a in activities:
            if "reference" in a:
                ref = a["reference"]
                if ref in pubs_by_id:
                    a["reference_title"] = pubs_by_id[ref]["title"]
        annotations = _annotations(obj)
        assocs = None
        if self.include_standard_annotations:
            std_to_activity = {}
            for a in activities:
                gene_id = a["gene"]
                activity_id = a.get("activity", None)
                if activity_id:
                    derived_standard_annotations.append((gene_id, activity_id))
                    std_to_activity[(gene_id, activity_id)] = a
                location_id = a.get("location", None)
                if location_id:
                    derived_standard_annotations.append((gene_id, location_id))
                    std_to_activity[(gene_id, location_id)] = a
                process_id = a.get("process", None)
                if process_id:
                    derived_standard_annotations.append((gene_id, process_id))
                    std_to_activity[(gene_id, process_id)] = a
            actual_gene_ids = []
            for k, v in lbl2id.items():
                if k in gene_ids:
                    actual_gene_ids.append(v)
            assocs = []
            adapter = get_adapter("amigo:")
            if not isinstance(adapter, AssociationProviderInterface):
                raise ValueError(f"Invalid adapter: {adapter}")
            # std_anns_accounted_for = set()
            for assoc in adapter.associations(actual_gene_ids):
                gene_obj = {
                    "id": assoc.subject,
                    "label": assoc.subject_label,
                }
                term_obj = {
                    "id": assoc.object,
                    "label": assoc.object_label,
                }
                gene_ref = _cls(gene_obj)
                term_ref = _cls(term_obj)
                ann_pubs = [str(x) for x in assoc.publications]
                ann_pubs = [p for p in ann_pubs if p.startswith("PMID") or p == "GO_REF:0000033"]
                assoc_obj = {
                    "gene": gene_ref,
                    "term": term_ref,
                    # "evidence": assoc.evidence_type,
                    "is_pub_in_gocam": len(set(ann_pubs).intersection(pmids)) > 0,
                }
                if ann_pubs:
                    assoc_obj["publication"] = ann_pubs[0]
                if (gene_ref, term_ref) in derived_standard_annotations:
                    assoc_obj["is_annotation_in_gocam"] = True
                    std_to_activity[(gene_ref, term_ref)]["has_standard_annotation"] = True
                assocs.append(assoc_obj)
            for a in activities:
                if "has_standard_annotation" not in a:
                    a["has_standard_annotation"] = False
        lbl2id = {k: v for k, v in lbl2id.items() if v.split(":")[0] not in ["ECO", "CARO", "obo"]}
        cam = {
            "id": id,
            "title": annotations["title"],
            "species": annotations.get("in_taxon", None),
            "provided_by": annotations.get("providedBy", None),
            "activities": activities,
            "publications": pubs,
            "idmap": lbl2id,
            "stats": {
                "num_genes": len(gene_ids),
                "num_processes": len(process_ids),
                "num_activities": len(activities),
                "num_relationships": len(relationships),
                "num_publications": len(pmids),
            },
        }
        if assocs:
            cam["associations"] = assocs
        return cam

    @property
    def label_adapter(self) -> BasicOntologyInterface:
        """Get the label adapter."""
        if self._label_adapter is None:
            self._label_adapter = get_adapter("sqlite:obo:envo")
        return self._label_adapter

    def _labelify(self, term: Dict):
        if "label" not in term or not term["label"]:
            term["label"] = self.label_adapter.label(term["id"])



"""
CurateGPT: A framework semi-assisted curation of knowledge bases.

Architecture
============

* :mod:`.store`: json object stores that allow for embedding based search
* :mod:`.wrappers`: wraps external APIs and data sources for ingest
* :mod:`.extract`: extraction of json objects from LLMs
* :mod:`.agents`: agents that chain together search and generate components
* :mod:`.formatters`: formats data objects for presentation to humans and machine agents
* :mod:`.app`: streamlit application


"""

import importlib_metadata

try:
    __version__ = importlib_metadata.version(__name__)
except importlib_metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover

from curategpt.extract import BasicExtractor, Extractor
from curategpt.store import ChromaDBAdapter, DBAdapter

__all__ = ["DBAdapter", "ChromaDBAdapter", "Extractor", "BasicExtractor"]



"""GO-CAM predictor class."""

import json
import logging
import sys
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Dict

import click
import yaml

from curategpt import BasicExtractor, DBAdapter
from curategpt.store import get_store
from curategpt.utils.eval_utils import Outcome, score_prediction
from curategpt.utils.llm_utils import query_model
from curategpt.wrappers.bio.gocam_wrapper import GOCAMWrapper

logger = logging.getLogger(__name__)


@dataclass
class GOCAMPredictor:

    store: DBAdapter = None
    database_type: str = field(default="chromadb")
    database_path: str = field(default="gocam")
    collection_name: str = field(default="gocam")
    model_name: str = field(default="gpt-4o")
    extractor: BasicExtractor = None
    include_standard_annotations: bool = False
    gocam_wrapper: GOCAMWrapper = field(default_factory=lambda: GOCAMWrapper())
    strict: bool = field(default=False)

    def __post_init__(self):
        self.store = get_store(self.database_type, self.database_path)
        self.extractor = BasicExtractor(model_name=self.model_name)
        self.gocam_wrapper = GOCAMWrapper(
            include_standard_annotations=self.include_standard_annotations
        )

    def gocam_by_id(self, id: str) -> Dict[str, Any]:
        return self.gocam_wrapper.object_by_id(id)
        # objs = list(self.store.find({"id": id}, collection=self.collection_name))
        # if len(objs) == 0:
        #    raise ValueError(f"GO-CAM model not found: {id}")
        # return objs[0][0]

    def predict_activity_unit(
        self, gocam: Dict[str, Any], stub: Dict[str, Any], num_examples=0
    ) -> Dict[str, Any]:
        """
        Predict the missing activity in a GO-CAM model.

        :param gocam: The GO-CAM model
        :param stub: The stub to fill in
        :param num_examples: The number of examples to generate
        :return: The predicted activity
        """
        model = self.extractor.model
        gocam_yaml = yaml.dump(gocam, sort_keys=False)
        examples = []
        if num_examples:
            for result, _score, _meta in self.store.search(
                gocam_yaml, limit=num_examples + 1, collection=self.collection_name
            ):
                if result["id"] == gocam["id"]:
                    continue
                examples.append(result)
            logger.info(f"Found {len(examples)} examples")

        stub_yaml = yaml.dump(stub, sort_keys=False)
        system = """
        You are an expert pathway curator. Your job is to provide a single
        additional activity for the provided GO-CAM model.
        """
        if examples:
            system += "Here are some examples other GO-CAM models:\n"
            for example in examples:
                system += f"""
                ```yaml
                {yaml.dump(example, sort_keys=False)}
                ```
                """
        prompt = f"""
        I will first provide the GO-CAM as YAML:

        ```yaml
        {gocam_yaml}
        ```

        Your job is to fill in this stub:

        ```yaml
        {stub_yaml}
        ```

        Provide YAML *ONLY* for this activity. DO NOT provide or correct
        YAML for existing activities. Try and include all relevant fields:
        gene, activity, process, and relationships.

        Example:

        ```yaml
        gene: <GeneName>
        activity: <MolecularActivityName>
        process: <BiologicalProcessName>
        location: <CellularLocationName>
        relationships:
          - type: <PredicateName>
            target_gene: <TargetGeneName>
            target_activity: <TargetMolecularActivityName>
        ```

        In your response, enclose the YAML in triple backticks.
        """
        logger.debug(f"System: {system}")
        logger.debug(f"Prompt: {prompt}")
        if self.model_name.startswith("gemini"):
            response = model.prompt(prompt=system + "\n===\n" + prompt)
        else:
            response = query_model(model, prompt=prompt, system=system)
            # response = model.prompt(prompt=prompt, system=system)
        text = response.text()
        toks = text.split("```")
        if len(toks) < 3:
            logger.error(f"Invalid response: {text}")
            return {}
        yaml_part = toks[1]
        if yaml_part.startswith("yaml"):
            yaml_part = yaml_part[4:]
        try:
            obj = yaml.safe_load(yaml_part)
        except Exception as e:
            obj = self.fix_yaml(yaml_part, e)
        try:
            Outcome(prediction=obj)
        except:
            logger.error(f"Invalid response: {obj}")
            return {}
        return obj

    def fix_yaml(self, yaml_text: str, e: Exception) -> str:
        """
        Fix the YAML text.

        :param yaml_text: The YAML text
        :return: The fixed YAML text
        """
        model = self.extractor.model
        response = model.prompt(
            system="""You provided this YAML but it doesn't parse. Return in JSON instead.
                       Only return valid JSON in payload, notyhing else""",
            prompt=f"```yaml\n{yaml_text}``` {e}",
        )
        text = response.text()
        if text.startswith("```json"):
            text = text[7:]
        if text.endswith("```"):
            text = text[:-3]
        try:
            obj = json.loads(text)
        except json.JSONDecodeError as e:
            if self.strict:
                raise ValueError(f"Invalid JSON response: {text}") from e
            else:
                obj = {}
        return obj


database_type_option = click.option(
    "--database_type", default="chromadb", help="The type of database to use"
)
database_path_option = click.option(
    "--database_path", default="gocam", help="The path to the database"
)
collection_name_option = click.option(
    "--collection_name", default="gocam", help="The name of the collection"
)
model_name_option = click.option("--model-name", default="gpt-4o", help="The name of the model")
num_examples_option = click.option(
    "--num-examples",
    "-N",
    type=click.INT,
    default=0,
    show_default=True,
    help="Number of RAG examples to generate",
)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
def main(verbose: int, quiet: bool):
    """
    CLI for GO-CAM prediction

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    # logger = logging.getLogger()
    logging.basicConfig()
    logger = logging.root
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)
    logger.info(f"Logger {logger.name} set to level {logger.level}")


@main.command()
@database_type_option
@database_path_option
@collection_name_option
@num_examples_option
@model_name_option
@click.option("--gocam-id", "-x", help="The GO-CAM model ID")
@click.option("--pmid", "-P", help="PMID to use for prediction")
@click.option("--output", "-o", help="Output file")
@click.option(
    "--include-standard-annotations/--no-include-standard-annotations",
    default=False,
    show_default=True,
    help="Include standard annotations",
)
@click.option("--prior-results-file", help="Previous results")
def predict(
    gocam_id,
    pmid,
    model_name,
    database_type,
    database_path,
    collection_name,
    num_examples,
    output,
    include_standard_annotations,
    prior_results_file,
):
    """
    Predict GO-CAMs.

    Example:

        gcpr -v  predict -N 5 --model-name claude-3-opus  -o results/gocam-run-claude3-opus-N5-run1.yaml

    We try and defend against API failures, e.g. implementing exponential backoff, but there is still
    potential for runs to fail, so we allow `--prior-results-file` to start off where we left off.

    Example (initial run):

        gcpr -v  predict -N 5 -o results/gocam-run-N5-run1.yaml

    After failure:

        gcpr -v  predict -N 5 --prior-results-file results/gocam-run-N5-run1.yaml -o results/gocam-run-N5-run1.yaml

    """
    already_done = []
    if prior_results_file:
        with open(prior_results_file, "r", encoding="utf-8") as f:
            already_done = [Outcome(**x).parameters["gocam_id"] for x in yaml.safe_load_all(f)]
            logger.info(f"Loaded {len(already_done)} already done")
    if output:
        output_file = open(output, "w", encoding="utf-8")
    else:
        output_file = sys.stdout
    predictor = GOCAMPredictor(
        database_type=database_type,
        database_path=database_path,
        collection_name=collection_name,
        model_name=model_name,
        include_standard_annotations=include_standard_annotations,
    )
    # set the list of GO-CAMs to predict; may be a specified singleton or ALL go-cams
    if gocam_id:
        gocam_ids = [gocam_id]
    else:
        if pmid:
            raise ValueError("PMID requires a GO-CAM ID")
        gocam_ids = [
            x["id"]
            for x, _s, _m in predictor.store.find({}, collection=collection_name, limit=9999)
        ]
        logger.info(f"Found {len(gocam_ids)} GO-CAM models")
    outcomes = []
    orig_pmid = pmid
    for gocam_id in gocam_ids:
        if gocam_id in already_done:
            logger.info(f"Skipping {gocam_id}, in already done list")
            continue
        original_gocam = predictor.gocam_by_id(gocam_id)
        title = original_gocam.get("title", None)
        test_gocam = deepcopy(original_gocam)
        if orig_pmid:
            pmids = [orig_pmid]
        else:
            pmids = [a["reference"] for a in test_gocam["activities"]]
        main_outcome = Outcome(prediction=[], expected=[])
        for pmid in pmids:
            refs = [a["reference"] for a in test_gocam["activities"]]
            # the gocam model used here is simplified in that each activity has a single ref
            test_gocam["activities"] = [
                a for a in original_gocam["activities"] if a["reference"] != pmid
            ]
            expected = [a for a in original_gocam["activities"] if a["reference"] == pmid]
            if not expected:
                raise ValueError(
                    f"Activity with reference {pmid} not found in GO-CAM model {gocam_id} // {refs}"
                )
            if len(test_gocam["activities"]) >= len(original_gocam["activities"]):
                raise ValueError(f"Invalid number of activities in GO-CAM model {gocam_id}")
            try:
                result = predictor.predict_activity_unit(
                    test_gocam, {"reference": pmid}, num_examples=num_examples
                )
            except Exception as e:
                logger.error(f"Encountered error: {e}")
                continue
            print("## PREDICTION")
            print(yaml.dump(result, sort_keys=False))
            print("## EXPECTED")
            print(yaml.dump(expected, sort_keys=False))
            outcome = score_prediction(result, expected)
            outcome.parameters = {
                "gocam_id": gocam_id,
                "title": title,
                "pmid": pmid,
                "num_examples": num_examples,
                "model_name": predictor.model_name,
                "include_standard_annotations": include_standard_annotations,
            }
            print("## OUTCOME")
            print(
                yaml.dump(
                    outcome.model_dump(
                        exclude={"predicted", "expected", "has_standard_annotation"}
                    ),
                    sort_keys=False,
                )
            )
            output_file.write("---\n")
            output_file.write(yaml.dump(outcome.model_dump(), sort_keys=False))
            output_file.write("\n")
            output_file.flush()
            outcomes.append(outcome)
    if len(outcomes) > 1:
        main_outcome.append_outcomes(outcomes)
        print("## MAIN OUTCOME")
        print(yaml.dump(main_outcome.model_dump(), sort_keys=False))
    output_file.close()


@main.command()
@click.option("--output", "-o", help="Output file")
@click.option("--tsv", help="TSV Output file")
@click.option("--add-species/--no-add-species", default=False, help="Add species to the output")
@click.argument("results_file")
def summarize_results(results_file, tsv, output, add_species):
    """
    Summarize the results of a previous run through gocam-predict.

    Takes the YAML output of the prediction step.
    """
    wrapper = GOCAMWrapper()
    species_by_gocam_id = {}

    def _get_species(id: str):
        if id not in species_by_gocam_id:
            gocam = wrapper.object_by_id(id)
            species = gocam.get("species", None)
            if species:
                species_by_gocam_id[id] = species
            else:
                species_by_gocam_id[id] = "unknown"
            logger.debug(f"Getting species for {id} => {species_by_gocam_id[id]}")
        return species_by_gocam_id[id]

    # read in results file YAML, flattening outcomes
    with open(results_file, "r", encoding="utf-8") as f:

        def _outcome(x: dict):
            outcome = Outcome(**x)
            params = outcome.parameters
            if add_species:
                params["species"] = _get_species(params["gocam_id"])
            logger.debug(f"Loaded outcome: {yaml.dump(outcome.model_dump)}")
            return outcome

        outcomes = [_outcome(x) for x in yaml.safe_load_all(f)]
    logger.info(f"Loaded {len(outcomes)} outcomes")
    rows = [outcome.flatten() for outcome in outcomes]
    if tsv:
        import pandas as pd

        df = pd.DataFrame(rows)
        df.to_csv(tsv, index=False, sep="\t")
        print(df.describe())
    if output:
        with open(output, "w", encoding="utf-8") as f:
            yaml.dump_all([outcome.model_dump() for outcome in outcomes], f, sort_keys=False)
    main_outcome = Outcome(prediction=[], expected=[])
    for outcome in outcomes:
        outcome.by_field = {}
        outcome.ixn_by_field = {}
    main_outcome.append_outcomes(outcomes)
    print(yaml.dump(main_outcome.model_dump(), sort_keys=False))


# add footer boilerplate for click
if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from curategpt import BasicExtractor, ChromaDBAdapter, DBAdapter, Extractor
from curategpt.agents.chat_agent import ChatResponse
from curategpt.app.cart import Cart
from curategpt.extract import AnnotatedObject


@dataclass
class PageState:
    predicted_object: AnnotatedObject = None
    chat_response: ChatResponse = None
    results: List = None
    selected: Any = None


@dataclass
class ApplicationState:
    page: Optional[str] = None
    db: DBAdapter = field(default_factory=ChromaDBAdapter)
    extractor: Extractor = field(default_factory=BasicExtractor)
    cart: Cart = field(default_factory=Cart)
    page_states: Dict[str, PageState] = field(default_factory=dict)
    # selected: Any = None

    def get_page_state(self, page_name: str) -> PageState:
        if page_name not in self.page_states:
            self.page_states[page_name] = PageState()
        return self.page_states[page_name]


def get_state(st) -> ApplicationState:
    """
    Gets the application state from the streamlit session state

    :param st:
    :return:
    """
    if "state" not in st.session_state:
        st.session_state["state"] = ApplicationState()
    return st.session_state["state"]



"""Chat with a KB."""

import json
import logging
import re
from dataclasses import dataclass
from enum import Enum
from random import shuffle
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import inflection
import yaml
from pydantic import BaseModel, ConfigDict

from curategpt.agents.base_agent import BaseAgent
from curategpt.formatters.format_utils import remove_formatting
from curategpt.store.db_adapter import SEARCH_RESULT
from curategpt.utils.tokens import estimate_num_tokens, max_tokens_by_model

logger = logging.getLogger(__name__)

PROMPT_TEMPLATE = """
Which of the following entities best matches the concept "{query}"?
---
{body}
---
Give the result as a json list of references, for example
[2, 3]. If there are no matches, return an empty list.
Return only matches where you are confident the match is the same.


Concept to match: "{query}"
Answer:
"""


class MappingPredicate(str, Enum):
    SAME_AS = "SAME_AS"
    CLOSE_MATCH = "CLOSE_MATCH"
    BROAD_MATCH = "BROAD_MATCH"
    NARROW_MATCH = "NARROW_MATCH"
    RELATED_MATCH = "RELATED_MATCH"
    DIFFERENT_FROM = "DIFFERENT_FROM"
    UNKNOWN = "UNKNOWN"


class Mapping(BaseModel):
    """Response from chat engine."""

    model_config = ConfigDict(protected_namespaces=())
    subject_id: str
    object_id: str
    predicate_id: Optional[MappingPredicate] = None


class MappingSet(BaseModel):

    model_config = ConfigDict(protected_namespaces=())
    mappings: List[Mapping]
    prompt: str = None
    response_text: str = None


@dataclass
class MappingAgent(BaseAgent):
    """
    An agent to map/align entities.
    """

    relevance_factor: float = 1.0
    """Relevance factor for diversifying search results using MMR.
    high is recommended for this task"""

    def match(
        self,
        query: Union[str, Dict[str, Any]],
        limit: int = None,
        randomize_order: bool = False,
        include_predicates: bool = False,
        fields: List[str] = None,
        id_field: str = "id",
        **kwargs,
    ) -> MappingSet:
        """
        Match entities

        :param query:
        :param limit:
        :param randomize_order: randomize the order in which candidates are presented (mostly for testing purposes)
        :param kwargs:
        :return:
        """
        # TODO: this approach doesn't work well - better do do the MapperGPT approach
        # and evaluate each mapping individually
        if self.extractor is None:
            raise ValueError("Extractor must be set.")
        # de-camelcase-ify if the query is camelCase
        if re.match(r".*[a-z][A-Z]", query):
            query = inflection.titleize(query).lower()
        if limit is None:
            limit = 10
        kb_results = list(
            self.knowledge_source.search(
                query, relevance_factor=self.relevance_factor, limit=limit, **kwargs
            )
        )
        if randomize_order:
            # primarily for testing
            shuffle(kb_results)
        if include_predicates:
            mappings = list(self.categorize_mappings(query, kb_results, **kwargs))
            return MappingSet(mappings=mappings)
        model = self.extractor.model
        while True:
            i = 0
            references = {}
            objects = {}
            texts = []
            current_length = 0
            for obj, _, _obj_meta in kb_results:
                i += 1
                obj_text = yaml.dump(
                    {k: v for k, v in obj.items() if v and (fields is None or k in fields)},
                    sort_keys=False,
                )
                references[str(i)] = obj_text
                objects[str(i)] = obj
                texts.append(f"## REF {i}\n{obj_text}")
                current_length += len(obj_text)
            prompt = PROMPT_TEMPLATE.format(body="".join(texts), query=query)
            logger.debug(f"Prompt: {prompt}")
            estimated_length = estimate_num_tokens([prompt])
            logger.debug(f"Max tokens {model.model_id}: {max_tokens_by_model(model.model_id)}")
            # TODO: use a more precise estimate of the length
            if estimated_length + 300 < max_tokens_by_model(model.model_id):
                break
            else:
                # remove least relevant
                if not kb_results:
                    raise ValueError(f"Prompt too long: {prompt}.")
                kb_results.pop()
        response = model.prompt(prompt)

        # Need to remove Markdown formatting here or it won't parse as JSON
        response_text = remove_formatting(text=response.text(), expect_format="json")

        mappings = []
        try:
            for m in json.loads(response_text):
                m = str(m)
                if m in objects:
                    object_id = objects.get(m)["id"]
                else:
                    object_id = m
                mappings.append(
                    Mapping(
                        subject_id=query,
                        object_id=object_id,
                        predicate=MappingPredicate.SAME_AS,
                        prompt=prompt,
                    )
                )
        except json.decoder.JSONDecodeError:
            # This will happen if the response is still not valid JSON
            # This returns an empty set of mappings, but the prompt and response text are retained
            return MappingSet(mappings=mappings, prompt=prompt, response_text=response_text)

        return MappingSet(mappings=mappings, prompt=prompt, response_text=response_text)

    def categorize_mappings(
        self, query: Union[str, Dict[str, Any]], kb_results: List[SEARCH_RESULT], **kwargs
    ) -> Iterator[Mapping]:
        """
        Categorize mappings predicate

        :param query:
        :param kb_results:
        :return:
        """
        for result in kb_results:
            prompt = (
                "I will give you two concepts, and your job is to tell me how they are related.\n\n"
            )
            prompt += f"Allowed answers are {', '.join(MappingPredicate.__members__)}:\n"
            prompt += "What is the relationship between these two concepts:\n"
            prompt += f"Concept A:\n{query}\n"
            prompt += f"Concept B:\n{yaml.dump(result[0], sort_keys=False)}"
            prompt += "\n\n"
            prompt += "Relationship:\n"
            model = self.extractor.model
            response = model.prompt(prompt)
            response_text = response.text()
            if response_text in MappingPredicate.__members__:
                pred = MappingPredicate[response_text]
            else:
                pred = MappingPredicate.UNKNOWN
            m = Mapping(subject_id=query, object_id=result[0]["id"], predicate_id=pred)
            yield m

    def find_links(self, other_collection: str) -> Iterator[Tuple[str, str, str]]:
        """
        Find links between elements in this collection and another collection

        :param other_collection:
        :return:
        """
        # TODO
        for _, _, info in self.knowledge_source.find(
            collection=other_collection, include=["embeddings", "documents", "metadatas"]
        ):
            embeddings = info["embeddings"]
            self.knowledge_source.find(embeddings, limit=10)



"""Chat with a KB."""

import logging
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Iterable, Iterator, List, Optional

import requests_cache
from oaklib import BasicOntologyInterface

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "https://rest.uniprot.org/uniprotkb"


@dataclass
class UniprotWrapper(BaseWrapper):
    """
    A wrapper over the UniProt API.
    """

    name: ClassVar[str] = "uniprot"

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "Protein"

    # taxon_id: str = field(default="NCBITaxon:9606")
    taxon_id: Optional[str] = None

    session: requests_cache.CachedSession = field(
        default_factory=lambda: requests_cache.CachedSession("uniprot")
    )

    def objects(
        self, collection: str = None, object_ids: Iterable[str] = None, **kwargs
    ) -> Iterator[Dict]:
        """
        All proteins
        """
        object_ids = object_ids or self.object_ids()
        return self.objects_by_ids(object_ids)

    def object_ids(self, taxon_id: str = None, **kwargs) -> Iterator[str]:
        """
        Get all gene ids for a given taxon id

        :param taxon_id:
        :param kwargs:
        :return:
        """
        url = f"{BASE_URL}/search"
        session = self.session
        taxon_id = taxon_id or self.taxon_id
        if not taxon_id:
            raise ValueError("Taxon ID is required")
        taxon_id = str(taxon_id).replace("NCBITaxon:", "")
        params = {
            "query": f"organism_id:{taxon_id} AND reviewed:true",
            # Query for E. coli using NCBI taxon ID and reviewed (Swiss-Prot) proteins
            "format": "json",  # Response format
            "size": 500,  # Number of results per page (max 500)
            "fields": "accession,id",  # Fields to retrieve
        }

        # Send the request
        logger.info(f"Getting proteins for taxon {taxon_id}")
        response = session.get(url, params=params)
        response.raise_for_status()
        data = response.json()
        entries = data.get("results", [])
        for entry in entries:
            logger.debug(f"Got entry: {entry}")
            yield entry["primaryAccession"]

    def objects_by_ids(self, object_ids: List[str]) -> List[Dict]:
        session = self.session
        objs = []
        for object_id in object_ids:
            if ":" in object_id:
                pfx, object_id = object_id.split(":")[1]
                if pfx.lower() not in ["uniprot", "uniprotkb"]:
                    raise ValueError(f"Invalid object id prefix: {pfx}")
            url = f"{BASE_URL}/{object_id}.json"
            logger.info(f"Getting protein data for {object_id} from {url}")
            response = session.get(url)
            response.raise_for_status()
            data = response.json()
            objs.append(data)
        return objs



"""
Adapters for different document stores.

* Base class: :class:`DBAdapter`

Currently only one implementation is provided, for ChromaDB.
In future this will index

- MongoDB
- ElasticSearch
- Solr
- Postgres
- SQLite

Note: this package may become an independent project called linkml-store
in the future.
"""

from .chromadb_adapter import ChromaDBAdapter
from .db_adapter import DBAdapter
from .duckdb_adapter import DuckDBAdapter
from .metadata import CollectionMetadata
from .schema_proxy import SchemaProxy

__all__ = [
    "DBAdapter",
    "ChromaDBAdapter",
    "DuckDBAdapter",
    "SchemaProxy",
    "CollectionMetadata",
    "get_store",
]


def get_all_subclasses(cls):
    """Recursively get all subclasses of a given class."""
    direct_subclasses = cls.__subclasses__()
    return direct_subclasses + [
        s for subclass in direct_subclasses for s in get_all_subclasses(subclass)
    ]


def get_store(name: str, *args, **kwargs) -> DBAdapter:  # duckdb_vss or chromadb
    from .in_memory_adapter import InMemoryAdapter  # noqa F401

    # noqa I005

    for c in get_all_subclasses(DBAdapter):
        if c.name == name:
            return c(*args, **kwargs)
    raise ValueError(f"Unknown view {name}, not found in {get_all_subclasses(DBAdapter)}")



"""Chat with a KB."""

import logging
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, Dict, Iterable, Iterator, List, Optional, Union

from deprecation import deprecated

from curategpt import ChromaDBAdapter
from curategpt.extract import Extractor
from curategpt.store import DBAdapter
from curategpt.store.db_adapter import SEARCH_RESULT

logger = logging.getLogger(__name__)


@dataclass
class BaseWrapper(ABC):  # noqa: B024
    """
    A virtual store that implements a view over some remote or external source.
    """

    source_locator: Optional[Union[str, Path]] = None

    local_store: DBAdapter = None
    """Adapter to local knowledge base used to cache results."""

    extractor: Extractor = None

    name: ClassVar[str] = "__dbview__"

    default_embedding_model = "openai:"

    default_object_type = "Publication"

    prefix: ClassVar[str] = None

    search_limit_multiplier: ClassVar[int] = 3

    max_text_length = 3000
    text_overlap = 200

    def search(
        self,
        text: str,
        collection: str = None,
        cache: bool = True,
        expand: bool = True,
        limit: Optional[int] = None,
        external_search_limit: Optional[int] = None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        """
        Search an external source and cache the results in the local store.

        :param text:
        :param collection: used for caching
        :param kwargs:
        :return:
        """
        logger.info(f"Searching for {text}")
        if external_search_limit is None and limit is not None:
            external_search_limit = limit * self.search_limit_multiplier
        parsed_data = self.external_search(
            text, expand=expand, limit=external_search_limit, **kwargs
        )
        db = self.local_store
        if db is None:
            tmpdir = Path("/tmp")
            db = ChromaDBAdapter(str(tmpdir))
        if collection is None:
            collection = self._cached_collection_name(is_temp=not cache)
        if not cache:
            db.remove_collection(collection, exists_ok=True)
        logger.info(f"Inserting {len(parsed_data)} records into {collection}")
        db.upsert(parsed_data, collection=collection, model=self.default_embedding_model)
        db.update_collection_metadata(
            collection,
            object_type="Publication",
            description=f"Special cache for {self.name} searches",
        )
        yield from db.search(text, collection=collection, limit=limit, **kwargs)

    def objects(
        self, collection: str = None, object_ids: Iterable[str] = None, **kwargs
    ) -> Iterator[Dict]:
        """
        Return all objects in the view.

        :param collection:
        :param object_ids: Optional list of IDs to fetch
        :param kwargs:
        :return:
        """
        if object_ids:
            yield from self.objects_by_ids(object_ids, **kwargs)
        else:
            raise NotImplementedError

    def objects_by_ids(self, object_ids: List[str]) -> List[Dict]:
        """
        Return objects by their IDs.

        :param object_ids:
        :return:
        """
        raise NotImplementedError

    def external_search(self, text: str, expand: bool = True, **kwargs) -> List[Dict]:
        """
        Search an external source and return the results.

        :param text:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def extract_concepts_from_text(self, text: str, **kwargs):
        model = self.extractor.model
        response = model.prompt(
            text, system="generate a semi-colon separated list of the most relevant terms"
        )
        terms = response.text().split(";")
        return terms

    @deprecated("Use a chat agent with a wrapper as source instead")
    def chat(
        self,
        query: str,
        collection: str = None,
        **kwargs,
    ) -> Any:
        """
        Chat with a wrapper view (e.g pubmed).

        :param query:
        :param collection:
        :param kwargs:
        :return:
        """
        from curategpt.agents.chat_agent import ChatAgent

        # prime the pubmed cache
        if collection is None:
            collection = self._cached_collection_name()
        logger.info(f"Ensure pubmed cached for {query}, kwargs={kwargs}, self={self}")
        _ = list(self.search(query, collection=collection, **kwargs))
        # ensure the collection exists and is configured correctly
        self.local_store.update_collection_metadata(
            collection,
            model=self.default_embedding_model,
            object_type=self.default_object_type,
            description=f"Special cache for {self.name} searches",
        )
        chat = ChatAgent(knowledge_source=self.local_store, extractor=self.extractor)
        response = chat.chat(query, collection=collection)
        return response

    def _cached_collection_name(self, is_temp=False) -> str:
        if is_temp:
            return f"__{self.name}_temp__"
        else:
            return self.name + "_api_cached"

    def split_objects(self, objects: List[Dict], text_field="text", id_field="id") -> List[Dict]:
        """
        Split objects with text above a certain length into multiple objects.

        :param objects:
        :return:
        """
        new_objects = []
        for obj in objects:
            if len(obj[text_field]) > self.max_text_length:
                obj_id = obj[id_field]
                text = obj[text_field]
                n = 0
                while text:
                    new_obj = obj.copy()
                    n += 1
                    new_obj[id_field] = f"{obj_id}#{n}"
                    new_obj[text_field] = text[: self.max_text_length + self.text_overlap]
                    new_objects.append(new_obj)
                    text = text[self.max_text_length :]
            else:
                new_objects.append(obj)
        return new_objects

    def create_curie(self, local_id: str) -> str:
        """
        Create a CURIE for the given local ID.

        :param local_id:
        :return:
        """
        return f"{self.prefix}:{local_id}"

    def wrap_object(self, obj: Dict) -> Iterator[Dict]:
        """
        Convert an object from the source representation to the store representation.

        :param obj:
        :return:
        """
        raise NotImplementedError

    def unwrap_object(self, obj: Dict[str, Any], store: DBAdapter, **kwargs) -> Any:
        """
        Convert an object from the store to the source representation.

        reverse transform of `as_object`

        :param obj:
        :param store:
        :param kwargs:
        :return:
        """
        raise NotImplementedError



"""Chat with a KB."""

import logging
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import requests_cache
from oaklib import BasicOntologyInterface

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "https://www.alliancegenome.org/api"


@dataclass
class AllianceGeneWrapper(BaseWrapper):
    """
    A wrapper over a Alliance (AGR) gene API.

    In future this may be extended to other objects.
    """

    name: ClassVar[str] = "alliance_gene"

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "Gene"

    taxon_id: str = field(default="NCBITaxon:9606")

    def object_ids(self, taxon_id: str = None, **kwargs) -> Iterator[str]:
        """
        Get all gene ids for a given taxon id

        :param taxon_id:
        :param kwargs:
        :return:
        """
        session = requests_cache.CachedSession("alliance")

        if not taxon_id:
            taxon_id = self.taxon_id
        response = session.get(
            f"{BASE_URL}/geneMap/geneIDs",
            params={
                "taxonID": taxon_id,
                "rows": 50000,
            },
        )
        response.raise_for_status()
        gene_ids = response.text.split(",")
        yield from gene_ids

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        """
        All genes

        :param collection:
        :param object_ids:
        :param kwargs:
        :return:
        """
        session = requests_cache.CachedSession("alliance")

        if not object_ids:
            gene_ids = self.object_ids(**kwargs)
        else:
            gene_ids = object_ids

        for gene_id in gene_ids:
            response = session.get(f"{BASE_URL}/gene/{gene_id}")
            if response.status_code == 400:
                # unknown gene
                continue
            response.raise_for_status()
            obj = response.json()
            yield obj



"""Chat with a KB."""

import json
import logging
import os
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import requests
import requests_cache

from curategpt.wrappers import BaseWrapper

URL = "https://api.fairsharing.org"
SIGNIN_URL = f"{URL}/users/sign_in"
FETCH_URL = f"{URL}/fairsharing_records"


logger = logging.getLogger(__name__)


MAX_ID = 10000


@dataclass
class FAIRSharingWrapper(BaseWrapper):
    """
    A wrapper over the NMDC Biosample API.
    """

    name: ClassVar[str] = "fairsharing"

    default_object_type = "Metadata"

    session: requests_cache.CachedSession = field(
        default_factory=lambda: requests_cache.CachedSession("fairsharing")
    )

    user: str = field(default=os.getenv("FAIRSHARING_USER"))

    password: str = field(default=os.getenv("FAIRSHARING_PASSWORD"))

    _jwt: str = None

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        exclude = [462]
        for id in range(1, MAX_ID):
            if id in exclude:
                continue
            obj = self.object_by_id(id)
            if obj:
                yield obj

    @property
    def jwt(self) -> str:
        if self._jwt is None:
            payload = {"user": {"login": self.user, "password": self.password}}
            headers = {"Accept": "application/json", "Content-Type": "application/json"}

            response = requests.request(
                "POST", SIGNIN_URL, headers=headers, data=json.dumps(payload)
            )

            # Get the JWT from the response.text to use in the next part.
            data = response.json()
            self._jwt = data["jwt"]

        return self._jwt

    def object_by_id(self, object_id: str) -> Optional[Dict]:

        url = f"{FETCH_URL}/{object_id}"

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": "Bearer {0}".format(self.jwt),
        }

        # sleep(0.05)
        response = self.session.request("GET", url, headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            logger.warning(f"Could not download records for {object_id} from {url}")
            return None



"""Wrapper for JSON (or YAML) documents."""

import json
import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

import yaml
from jsonpath_ng import parse

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


@dataclass
class JSONWrapper(BaseWrapper):
    """
    A wrapper over a json (or yaml) document.

    Uses json path expressions

    This is a static wrapper: it cannot be searched
    """

    name: ClassVar[str] = "json"
    format: str = None
    path_expression: str = None

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        path = self.source_locator

        with open(path) as file:
            print(f"Loading {path}")
            format = self.format or str(path).split(".")[-1]
            if format == "json":
                obj = json.load(file)
            elif format == "yaml":
                obj = yaml.safe_load(file)
            else:
                raise ValueError(f"Unknown format {format}")
        yield from self.wrap_object(obj)

    def wrap_object(self, obj: Dict) -> Iterator[Dict]:
        jsonpath_expression = parse(self.path_expression)
        for match in jsonpath_expression.find(obj):
            yield match.value



from typing import Any, Dict, Iterator, List, Optional, Set, Tuple

from pydantic import BaseModel, ConfigDict

SEARCH_RESULT = Tuple[Dict[str, Any], Dict, float, Optional[Dict]]


class DuckDBSearchResult(BaseModel):

    model_config = ConfigDict(protected_namespaces=())

    ids: Optional[str] = None
    metadatas: Optional[Dict[str, Any]] = None
    embeddings: Optional[List[float]] = None
    documents: Optional[str] = None
    distances: Optional[float] = 0  # TODO: technically this is simple cosim similarity
    include: Optional[Set[str]] = None

    def to_json(self, indent: int = 2):
        return self.model_dump_json(include=self.include, indent=indent)

    def to_dict(self):
        if self.include:
            return self.model_dump(include=self.include)
        return self.model_dump()

    def __repr__(self, indent: int = 2):
        return self.model_dump_json(include=self.include, indent=indent)

    def __iter__(self) -> Iterator[SEARCH_RESULT]:
        # TODO vocab.py for 'VARIABLES', but circular import
        metadata_include = "metadatas" in self.include if self.include else True
        embeddings_include = "embeddings" in self.include if self.include else True
        documents_include = "documents" in self.include if self.include else True
        similarity_include = "distances" in self.include if self.include else True

        obj = self.metadatas if metadata_include else {}
        meta = {
            "_embeddings": self.embeddings if embeddings_include else None,
            "documents": self.documents if documents_include else None,
        }
        distance = self.distances if similarity_include else None

        yield obj, distance, meta



import logging
from dataclasses import dataclass
from typing import List

from curategpt.agents.base_agent import BaseAgent
from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)


@dataclass
class SummarizationAgent(BaseAgent):
    """
    An agent to summarize entities

    AKA SPINDOCTOR/TALISMAN
    """

    def summarize(
        self,
        object_ids: List[str],
        description_field: str,
        name_field: str,
        strict: bool = False,
        system_prompt: str = None,
    ):
        """
        Summarize a list of objects.

        Example:
        -------
        >>> extractor = BasicExtractor()
        >>> wrapper = get_wrapper("alliance_gene")
        >>> agent = SummarizationAgent(wrapper, extractor=extractor)
        >>> gene_ids = ["HGNC:9221", "HGNC:11195", "HGNC:6348", "HGNC:7553"]
        >>> response = agent.summarize(
        ...               gene_ids,
        ...               name_field="symbol",
        ...               description_field="automatedGeneSynopsis",
        ...               system_prompt="What function do these genes have in common?",
        ...           )
        >>> print(response)

        :param object_ids:
        :param description_field:
        :param name_field:
        :param strict:
        :param system_prompt:
        :return:
        """
        if not description_field:
            raise ValueError("Must provide a description field")
        if not name_field:
            raise ValueError("Must provide a name field")
        knowledge_source = self.knowledge_source
        if isinstance(knowledge_source, BaseWrapper):
            object_iter = knowledge_source.objects(self.knowledge_source_collection, object_ids)
        else:
            object_iter = knowledge_source.lookup_multiple(
                object_ids, collection=self.knowledge_source_collection
            )
        objects = list(object_iter)
        descriptions = [
            (obj.get(name_field, ""), obj.get(description_field, None)) for obj in objects
        ]
        if any(desc[0] is None for desc in descriptions):
            raise ValueError(f"Missing name for objects: {objects}")
        if strict:
            missing_descriptions = [name for name, desc in descriptions if desc is None]
            if missing_descriptions:
                raise ValueError(f"Missing descriptions for objects: {missing_descriptions}")
        if not descriptions:
            raise ValueError(f"No descriptions found for objects: {objects}")
        desc_lengths = [len(desc[1] or "") for desc in descriptions]
        max_desc_length = max(desc_lengths)
        if max_desc_length > 30:
            sep = "\n---\n"
        else:
            sep = "; "
        text = sep.join([f"{desc[0]}: {desc[1]}" for desc in descriptions])
        logger.info(f"Summarizing {len(descriptions)} objects with {len(text)} characters")
        model = self.extractor.model
        if not system_prompt:
            system_prompt = "Summarize these entities"
        response = model.prompt(text, system=system_prompt)
        return response.text()



import logging
import platform
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import TextIO

import yaml

from curategpt import BasicExtractor, ChromaDBAdapter
from curategpt.agents.dragon_agent import DragonAgent
from curategpt.evaluation.dae_evaluator import DatabaseAugmentedCompletionEvaluator
from curategpt.evaluation.evaluation_datamodel import Task
from curategpt.evaluation.splitter import stratify_collection_to_store

logger = logging.getLogger(__name__)


def run_task(
    task: Task,
    report_path=None,
    report_file: TextIO = None,
    report_tsv_file: TextIO = None,
    fresh=False,
    **kwargs,
) -> Task:
    """
    Evaluate the agent on a test collection.

    :param task:
    :param report_path:
    :param report_file:
    :param fresh: if True, overwrite existing results file
    :param kwargs: passed to the evaluator
    :return: the task with results
    """
    task = deepcopy(task)
    if not task.working_directory:
        raise ValueError("Working directory must be specified")
    wd = Path(task.working_directory)
    wd.mkdir(exist_ok=True, parents=True)
    results_file_path = wd / f"{task.id}.results.yaml"
    if results_file_path.exists():
        logger.info(f"Results file exists at {results_file_path}")
        if not fresh:
            logger.info(f"Loading results from {results_file_path}")
            task = Task.parse_obj(yaml.safe_load(results_file_path.open()))
            return task
        else:
            logger.info("Overwriting existing results file...")
    extractor = BasicExtractor(model_name=task.model_name)
    if task.target_db_path is None:
        task.target_db_path = str(wd / "db")
    target_path = task.target_db_path
    if not target_path:
        raise ValueError("Target database path must be specified")
    logger.info(f"Stratifying collection {task.source_collection} from {task.source_db_path}")
    db = ChromaDBAdapter(task.source_db_path)
    if task.stratified_collection:
        sc = task.stratified_collection
        sc_dict = {
            "training": sc.training_set_collection,
            "testing": sc.testing_set_collection,
        }
        logger.debug(f"Stratified collection: {sc}")
    else:
        sc_dict = stratify_collection_to_store(
            db,
            task.source_collection,
            output_path=target_path,
            num_training=task.num_training,
            num_testing=task.num_testing,
            num_validation=task.num_validation,
            embedding_model=task.embedding_model_name,
            force=fresh,
        )
    tdb = ChromaDBAdapter(target_path)
    # set start time to current time (ISO format)
    task.task_started = str(datetime.now())
    # get current operating system
    task.executed_on = (
        f"{platform.system()}-{platform.release()}-{platform.version()}-{platform.machine()}"
    )
    agent = DragonAgent(knowledge_source=tdb, knowledge_source_collection="", extractor=extractor)
    if task.additional_collections:
        if len(task.additional_collections) > 1:
            raise NotImplementedError("Only one additional collection is supported")
        agent.document_adapter = tdb
        agent.document_adapter_collection = task.additional_collections[0]
    # TODO: use the task object directly in the evaluator
    evaluator = DatabaseAugmentedCompletionEvaluator(
        agent=agent, fields_to_predict=task.fields_to_predict, fields_to_mask=task.fields_to_mask
    )
    if report_path is not None:
        task.report_path = report_path
    if task.report_path is not None:
        report_file = open(task.report_path, "w")
    if report_file is None:
        report_file = open(wd / f"{task.id}.log.yaml", "w")
    report_file.write("## Task\n")
    report_file.write(yaml.dump(task.dict(), sort_keys=False))
    if report_tsv_file is None:
        report_tsv_file = open(wd / f"{task.id}.results.tsv", "w")
        commented_yaml = yaml.dump(task.dict(), sort_keys=False)
        lines = [f"# {line}\n" for line in commented_yaml.splitlines()]
        report_tsv_file.write("".join(lines))
    logger.info(f"Evaluating agent {agent} on collection {sc_dict['testing']}")
    results = evaluator.evaluate(
        test_collection=sc_dict["testing"],
        num_tests=task.num_testing,
        collection=sc_dict["training"],
        report_file=report_file,
        report_tsv_file=report_tsv_file,
        generate_background=task.generate_background,
        **kwargs,
    )
    logger.info("Collecting results...")
    task.results = results
    # set finish time to current time (ISO format)
    task.task_finished = str(datetime.now())
    with open(results_file_path, "w") as file:
        file.write(yaml.dump(task.dict(), sort_keys=False))
    return task



"""Streamlit app for CurateGPT."""

import json
import logging
from typing import List, Union

import numpy as np
import streamlit as st
import yaml
from scipy.spatial import distance_matrix

from curategpt import BasicExtractor
from curategpt.agents import MappingAgent
from curategpt.agents.bootstrap_agent import BootstrapAgent, KnowledgeBaseSpecification
from curategpt.agents.chat_agent import ChatAgent, ChatResponse
from curategpt.agents.dase_agent import DatabaseAugmentedStructuredExtraction
from curategpt.agents.dragon_agent import DragonAgent
from curategpt.agents.evidence_agent import EvidenceAgent
from curategpt.app.components import (
    DimensionalityReductionOptions,
    limit_slider_component,
    vectors_to_fig,
)
from curategpt.app.helper import get_applicable_examples, get_case_collection
from curategpt.app.state import get_state
from curategpt.extract import OpenAIExtractor, RecursiveExtractor
from curategpt.wrappers import BaseWrapper
from curategpt.wrappers.investigation.jgi_wrapper import JGIWrapper
from curategpt.wrappers.literature import WikipediaWrapper
from curategpt.wrappers.literature.pubmed_wrapper import PubmedWrapper

PUBMED = "PubMed (via API)"
WIKIPEDIA = "Wikipedia (via API)"
JGI = "JGI (via API)"

CHAT = "Chat"
EXTRACT = "Extract"
SEARCH = "Search"
CLUSTER_SEARCH = "Cluster Search"
MATCH = "Match"
BOOTSTRAP = "Bootstrap"
CURATE = "Curate"
ADD_TO_CART = "Add to Cart"
# EXTRACT = "Extract"
CITESEEK = "CiteSeek"
CART = "Cart"
HELP = "Help"
EXAMPLES = "Examples"
ABOUT = "About"

NO_BACKGROUND_SELECTED = "No background collection"

MODELS = [
    "gpt-4o",
    "gpt-3.5-turbo",
    "gpt-4-turbo",
    "gpt-4",
    "chatgpt-16k",
    "nous-hermes-13b",
    "llama2",
]

logger = logging.getLogger(__name__)

state = get_state(st)
db = state.db
cart = state.cart


st.title("CurateGPT! _alpha_")
if not db.list_collection_names():
    st.warning("No collections found. Please use command line to load one.")

PAGES = [
    SEARCH,
    CLUSTER_SEARCH,
    CHAT,
    CURATE,
    EXTRACT,
    CITESEEK,
    MATCH,
    BOOTSTRAP,
    CART,
    ABOUT,
    HELP,
    EXAMPLES,
]


def _clear_sticky_page():
    logger.error("Clearing sticky page")
    state.page = None


# Sidebar with operation selection
option_selected = st.sidebar.selectbox(
    "Choose operation",
    PAGES,
    on_change=_clear_sticky_page,
)
option = state.page or option_selected
logger.error(f"Selected {option_selected}; sp={state.page}; opt={option}")
# logger.error(f"State: {state}")


def filtered_collection_names() -> List[str]:
    return [c for c in db.list_collection_names() if not c.endswith("_cached")]


collection = st.sidebar.selectbox(
    "Choose collection",
    filtered_collection_names() + [WIKIPEDIA, PUBMED, JGI],
    help="""
    A collection is a knowledge base. It could be anything, but
    it's likely your instance has some bio-ontologies pre-loaded.
    Select 'About' to see details of each collection
    """,
)

model_name = st.sidebar.selectbox(
    "Choose model",
    MODELS,
    help="""
    If this instance allows GPT-4 this likely works the best.
    (be considerate if someone else is paying).
    Open models may not do as well at extraction tasks
    (and they may be very slow).
    Note: if your instance is on EC2 it's likely open models
    that are not API backed will be unavailable or broken.
    """,
)

extraction_strategy = st.sidebar.selectbox(
    "Extraction strategy",
    ["Basic", "OAI Function", "SPIRES"],
    help="""
    The extraction strategy determines how the system will
    extract information from text.
    Note that both OpenAI functions and SPIRES require
    a schema to have been pre-loaded.
    """,
)

if extraction_strategy == "Basic":
    extractor = BasicExtractor()
elif extraction_strategy == "OAI Function":
    extractor = OpenAIExtractor()
elif extraction_strategy == "SPIRES":
    extractor = RecursiveExtractor()
else:
    raise ValueError(f"Unknown extraction strategy {extraction_strategy}")
state.extractor = extractor


background_collection = st.sidebar.selectbox(
    "Background knowledge",
    [NO_BACKGROUND_SELECTED, PUBMED, WIKIPEDIA, JGI] + list(db.list_collection_names()),
    help="""
    Background databases can be used to give additional context to the LLM.
    A standard pattern is to have a structured knowledge base as the main
    collection (this is used to find example records), and an unstructured
    database (e.g. github issues, abstracts, pdfs, ...) as background.
    Note you cannot currently add new databases using the UI. Contact
    the site admin to add new sources.
    """,
)

st.sidebar.markdown(f"Cart: {cart.size} items")

st.sidebar.markdown("Developed by the Monarch Initiative")


def get_chat_agent() -> Union[ChatAgent, BaseWrapper]:
    knowledge_source_collection = None
    if collection == PUBMED:
        source = PubmedWrapper(local_store=db, extractor=extractor)
    elif collection == WIKIPEDIA:
        source = WikipediaWrapper(local_store=db, extractor=extractor)
    elif collection == JGI:
        source = JGIWrapper(local_store=db, extractor=extractor)
    else:
        source = db
        knowledge_source_collection = collection
    return ChatAgent(
        knowledge_source=source,
        knowledge_source_collection=knowledge_source_collection,
        extractor=extractor,
    )


def ask_chatbot(query, expand=False) -> ChatResponse:
    return get_chat_agent().chat(query, expand=expand)


def html_table(rows: List[dict]) -> str:
    if len(rows) == 0:
        rows = [{"No data": "No data"}]
    hdr = rows[0].keys()
    html_content = '<table border="1">'
    cols = [f"<th>{h}</th>" for h in hdr]
    html_content += f"<tr>{''.join(cols)}</tr>"
    for row in rows:
        html_content += "<tr>"
        for col in hdr:
            v = row.get(col, "")
            if isinstance(v, dict):
                v = f"<pre>{yaml.dump(v, sort_keys=False)}</pre>"
            html_content += f"<td>{v}</td>"
        html_content += "</tr>"
    html_content += "</table>"
    return html_content


if option == CURATE:
    page_state = state.get_page_state(CURATE)
    st.subheader(f"Curate new object in {collection}")
    if page_state.selected is None:
        fields = db.field_names(collection=collection)
    else:
        fields = list(page_state.selected.keys())
    input_obj = {}
    for f in fields:
        v = ""
        if page_state.selected is not None:
            v = page_state.selected.get(f, "")
        input_obj[f] = st.text_input(f"{f}", value=v)

    extractor.model_name = model_name

    if page_state.selected is not None:
        if st.button("Clear"):
            page_state.selected = None
            st.success("Current Selection Cleared!")

    generate_background = st.checkbox(
        "Generate background",
        help="""
                                            If checked, a full text description is first generated from the LLM.
                                            This is then used as background knowledge to generate the target object.
                                            """,
    )
    instructions = st.text_input(
        "Additional Instructions",
        help="""
                                        Enter any additional instructions for the model here.
                                        E.g. 'You MUST include a definition field in your answer.
                                        """,
    )

    examples_limit = limit_slider_component()

    if st.button("Suggest"):
        daca = DragonAgent(knowledge_source=db, extractor=extractor)
        if background_collection != NO_BACKGROUND_SELECTED:
            # TODO: DRY
            if background_collection == PUBMED:
                daca.document_adapter = PubmedWrapper(local_store=db, extractor=extractor)
                daca.collection = None
            elif background_collection == WIKIPEDIA:
                daca.document_adapter = WikipediaWrapper(local_store=db, extractor=extractor)
                daca.collection = None
            else:
                daca.document_adapter = db
                daca.document_adapter_collection = background_collection
        st.write(f"Generating using: **{extractor.model_name}** using *{collection}* for examples")
        if background_collection:
            st.write(f"Using background knowledge from: *{background_collection}*")
        rules = [instructions] if instructions else None
        page_state.predicted_object = daca.complete(
            input_obj,
            generate_background=generate_background,
            collection=collection,
            rules=rules,
            limit=examples_limit,
        )

    if page_state.predicted_object:
        created = page_state.predicted_object
        obj = created.object
        st.subheader("Created object")
        # st.write("<pre>" + yaml.dump(obj, sort_keys=False) + "</pre>", unsafe_allow_html=True)
        st.code(yaml.dump(obj, sort_keys=False), language="yaml")
        add_button = st.button(f"Add to {collection}")
        if add_button:
            db.insert([obj], collection=collection)
            st.write("Added!!!")
        st.subheader("Debug info")
        st.write("Prompt:")
        st.code(created.annotations["prompt"])
        # st.write("Property:", property_query)


elif option == MATCH:
    page_state = state.get_page_state(MATCH)
    st.subheader(f"Match to entities in *{collection}*")
    search_query = st.text_input(
        "Match text",
        help="Enter label of concept to match.",
    )
    relevant_fields = st.text_input(
        "Relevant fields",
        help="Comma-separated (e.g. label, definition).",
    )
    limit = st.slider(
        "Max results",
        min_value=0,
        max_value=200,
        value=10,
        step=1,
        help="""
                                       Number of results max
                                       """,
    )

    if st.button("Match"):
        cm = db.collection_metadata(collection, include_derived=True)
        st.write(f"Searching over {cm.object_count} objects using embedding model {cm.model}")
        mapper = MappingAgent(knowledge_source=db, extractor=extractor)
        if not relevant_fields:
            relevant_fields = "label"
        relevant_fields = [f.strip() for f in relevant_fields.split(",")]
        results = mapper.match(
            search_query,
            collection=collection,
            fields=relevant_fields,
            limit=limit,
        )

        rows = [{"subject": search_query, "object": m.object_id} for m in results.mappings]
        html = html_table(rows)
        st.write(html, unsafe_allow_html=True)
        st.subheader("Prompt", help="for debugging")
        st.write(results.prompt)
        st.subheader("Response", help="for debugging")
        st.write(results.response_text)

# Search operation
elif option == SEARCH:
    page_state = state.get_page_state(SEARCH)
    st.subheader(f"Search documents in *{collection}*")
    search_query = st.text_input(
        "Search by text",
        help="Enter any text - embedding similarity will be used to find similar objects.",
    )

    relevance_factor = st.slider(
        "Relevance Factor",
        min_value=0.0,
        max_value=1.0,
        value=1.0,
        step=0.05,
        help="""
                                 How much to weight the relevance vs diversity of the search query.
                                 If this is set to less than 1.0, then MMR will be used to diversify the results.
                                 (this corresponds to the lambda parameter in the MMR formula)
                                 """,
    )

    # if not hasattr(st.session_state, "add_to_cart_index"):
    #    st.session_state.add_to_cart_index = None

    # if False and st.session_state.add_to_cart_index is not None:
    #    # add object from first result
    #    obj = page_state.results[st.session_state.add_to_cart_index][0]
    #    cart.add(obj)
    #    page_state.selected = obj
    #    st.session_state.add_to_cart_index = None  # Reset
    #    st.success("Document added to cart!!!!!")

    # elif st.button("Search"):
    if st.button("Search"):
        results = db.search(
            search_query, collection=collection, relevance_factor=relevance_factor, include=["*"]
        )
        page_state.results = list(results)
        st.session_state.add_to_cart_index = None  # Reset

    if page_state.results:
        results = page_state.results
        cm = db.collection_metadata(collection, include_derived=True)
        if cm:
            st.write(f"Searching over {cm.object_count} objects using embedding model {cm.model}")
        else:
            st.write(f"Dynamic search over {collection}...")

        def _flat(obj: dict, limit=40) -> dict:
            if not obj:
                return {}
            return {
                k: str(json.dumps(v) if isinstance(v, (list, dict)) else v)[0:limit]
                for k, v in obj.items()
            }

        rows = [
            {"rank": i + 1, "distance": distance, **_flat(obj), "doc": _flat(doc)}
            for i, (obj, distance, doc) in enumerate(results)
        ]
        html = html_table(rows)
        st.write(html, unsafe_allow_html=True)

        for i, (obj, _distance, _doc) in enumerate(results):
            st.write(f"## Result {i+1}")
            st.code(yaml.dump(obj, sort_keys=False))
            if st.button(f"Add to cart {i+1}"):
                cart.add(obj)
                st.success("Document added to cart!")
            if st.button(f"Curate {i+1}"):
                autoadd_page_state = state.get_page_state(CURATE)
                autoadd_page_state.selected = obj
                st.success("Curating!")
                state.page = CURATE
                st.experimental_rerun()
            if st.button(f"Find evidence {i+1}"):
                autoadd_page_state = state.get_page_state(CITESEEK)
                autoadd_page_state.selected = obj
                st.success("OK!")
                state.page = CITESEEK
                st.experimental_rerun()


elif option == CLUSTER_SEARCH:
    page_state = state.get_page_state(CLUSTER_SEARCH)
    st.subheader(f"Cluster Search documents in *{collection}*")
    search_query = st.text_input(
        "Search by text",
        help="Enter any text - embedding similarity will be used to find similar objects.",
    )

    relevance_factor = st.slider(
        "Relevance Factor",
        min_value=0.0,
        max_value=1.0,
        value=1.0,
        step=0.05,
        help="""How much to weight the relevance vs diversity of the search query.
                                 If this is set to less than 1.0, then MMR will be used to diversify the results.
                                 (this corresponds to the lambda parameter in the MMR formula)
                                 """,
    )

    limit = st.slider(
        "Max results",
        min_value=0,
        max_value=1000,
        value=50,
        step=1,
        help="""
                                   Number of results max
                                   """,
    )

    method = st.radio(
        "Dimensionality Reduction Method", options=list(DimensionalityReductionOptions), index=0
    )

    if st.button("Search"):
        cm = db.collection_metadata(collection, include_derived=True)
        st.write(f"Searching over {cm.object_count} objects using embedding model {cm.model}")
        include = ["*"]
        results = db.search(
            search_query,
            collection=collection,
            relevance_factor=relevance_factor,
            include=include,
            limit=limit,
        )
        labels = []
        vectors = []
        label_field = db.label_field(collection)
        id_field = db.identifier_field(collection)
        for i, (obj, _distance, doc) in enumerate(results):
            # use label field preferentially, then id field, then numeric
            labels.append(obj.get(label_field, obj.get(id_field, f"Object {i}")))
            vectors.append(np.array(doc["embeddings"]))
        distances = distance_matrix(vectors, vectors)
        fig = vectors_to_fig(labels, np.array(vectors), method=method)
        st.pyplot(fig)

elif option == EXTRACT:
    page_state = state.get_page_state(EXTRACT)
    st.subheader("Extract object", help="Extract a structured object from text.")
    st.write(f"Examples will be drawn from **{collection}**")
    if background_collection != NO_BACKGROUND_SELECTED:
        st.write(f"Background knowledge will be drawn from **{background_collection}**")
    text_query = st.text_input(
        "Text", help="Enter the label or description of the entity type you want to add."
    )
    generate_background = st.checkbox(
        "Generate background",
        help="""
                                        If checked, a full text description is first generated from the LLM.
                                        This is then used as background knowledge to generate the target object.
                                        """,
    )
    instructions = st.text_input(
        "Additional Instructions",
        help="""
                                    Enter any additional instructions for the model here.
                                    E.g. 'You MUST include a definition field in your answer.
                                    """,
    )

    examples_limit = limit_slider_component()

    examples = get_applicable_examples(collection, EXTRACT)
    extractor.model_name = model_name

    if st.button(EXTRACT):
        dase = DatabaseAugmentedStructuredExtraction(knowledge_source=db, extractor=extractor)
        if background_collection != NO_BACKGROUND_SELECTED:
            if background_collection == PUBMED:
                dase.document_adapter = PubmedWrapper(local_store=db, extractor=extractor)
                dase.collection = None
            elif background_collection == WIKIPEDIA:
                dase.document_adapter = WikipediaWrapper(local_store=db, extractor=extractor)
                dase.collection = None
            else:
                dase.document_adapter = db
                dase.document_adapter_collection = background_collection
        st.write(f"Generating using: **{extractor.model_name}** using *{collection}* for examples")
        if background_collection:
            st.write(f"Using background knowledge from: *{background_collection}*")
        rules = [instructions] if instructions else None
        page_state.predicted_object = dase.extract(
            text_query,
            generate_background=generate_background,
            collection=collection,
            rules=rules,
            limit=examples_limit,
        )

    if page_state.predicted_object:
        created = page_state.predicted_object
        obj = created.object
        st.subheader("Created object")
        # st.write("<pre>" + yaml.dump(obj, sort_keys=False) + "</pre>", unsafe_allow_html=True)
        st.code(yaml.dump(obj, sort_keys=False), language="yaml")
        add_button = st.button(f"Add to {collection}")
        if add_button:
            db.insert([obj], collection=collection)
            st.write("Added!!!")
        st.subheader("Debug info")
        st.write("Prompt:")
        st.code(created.annotations["prompt"])


elif option == CHAT:
    page_state = state.get_page_state(CHAT)
    st.subheader("Chat with a knowledge base")
    query = st.text_area(
        f"Ask me anything (within the scope of {collection})!",
        help="You can query the current knowledge base using natural language.",
    )

    limit = st.slider(
        "Detail",
        min_value=0,
        max_value=30,
        value=10,
        step=1,
        help="""
                                   Behind the scenes, N entries are fetched from the knowledge base,
                                   and these are fed to the LLM. Selecting more examples may give more
                                   complete results, but may also exceed context windows for the model.
                                   """,
    )
    expand = st.checkbox(
        "Expand query",
        help="""
                                                If checked, perform query expansion (pubmed only).
                                                """,
    )
    extractor.model_name = model_name
    examples = get_applicable_examples(collection, CHAT)
    st.write("Examples:")
    st.write(f"<details>{html_table(examples)}</details>", unsafe_allow_html=True)

    if st.button(CHAT):
        response = ask_chatbot(query, expand=expand)
        page_state.chat_response = response

    if page_state.chat_response:
        response = page_state.chat_response
        st.markdown(response.formatted_body)
        add_button = st.button("Add to your cart")
        if add_button:
            logger.error("Adding to cart")
            cart.add(response)
            st.write("Added to cart!")

        st.markdown("## References")
        for ref, text in response.references.items():
            st.subheader(f"Reference {ref}", anchor=f"ref-{ref}")
            st.code(text, language="yaml")
            if st.button(f"Add to cart {ref}"):
                # TODO: unpack
                cart.add({"text": text, "id": ref})
                st.success("Document added to cart!")
        if response.uncited_references:
            st.markdown("## Uncited references")
            st.caption(
                "These references were flagged as potentially relevant, but a citation was not detected."
            )
            for ref, text in response.uncited_references.items():
                st.subheader(f"Reference {ref}", anchor=f"ref-{ref}")
                st.code(text, language="yaml")

elif option == CITESEEK:
    page_state = state.get_page_state(CITESEEK)
    st.subheader("Find citations for a claim")
    v = None
    if page_state.selected is not None:
        v = yaml.dump(page_state.selected, sort_keys=False)
    query = st.text_area(
        f"Enter YAML object to be verified by {collection}",
        value=v,
        help="Copy the YAML from some of the other outputs of this tool.",
    )

    limit = st.slider(
        "Detail",
        min_value=0,
        max_value=30,
        value=10,
        step=1,
        help="""
                                   Behind the scenes, N entries are fetched from the knowledge base,
                                   and these are fed to the LLM. Selecting more examples may give more
                                   complete results, but may also exceed context windows for the model.
                                   """,
    )
    extractor.model_name = model_name
    # examples = get_applicable_examples(collection, CITESEEK)
    # st.write("Examples:")
    # st.write(f"<details>{html_table(examples)}</details>", unsafe_allow_html=True)

    if page_state.selected is not None:
        if st.button("Clear"):
            page_state.selected = None
            st.success("Current Selection Cleared!")

    if st.button(CITESEEK):
        chat_agent = get_chat_agent()
        ea = EvidenceAgent(chat_agent=chat_agent)
        try:
            query_obj = yaml.safe_load(query)
        except yaml.YAMLError:
            try:
                query_obj = json.loads(query)
            except json.JSONDecodeError as exc:
                st.warning(f"Invalid YAML or JSON: {exc}")
                query_obj = None
        if query_obj:
            response = ea.find_evidence(query_obj)
            # TODO: reuse code for this
            st.markdown(response.formatted_body)
            st.markdown("## References")
            for ref, text in response.references.items():
                st.subheader(f"Reference {ref}", anchor=f"ref-{ref}")
                st.code(text, language="yaml")
            if response.uncited_references:
                st.markdown("## Uncited references")
                st.caption(
                    "These references were flagged as potentially relevant, but a citation was not detected."
                )
                for ref, text in response.uncited_references.items():
                    st.subheader(f"Reference {ref}", anchor=f"ref-{ref}")
                    st.code(text, language="yaml")

elif option == BOOTSTRAP:
    page_state = state.get_page_state(BOOTSTRAP)
    st.subheader("Generate a schema and data for a new knowledge base.")

    extractor = BasicExtractor()
    extractor.model_name = model_name
    bootstrap_agent = BootstrapAgent(extractor=extractor)

    kb_name = st.text_input("KB Name", help="Name of the knowledge base, without spaces (e.g. 'ice_cream_kb')")
    description = st.text_input(
        "Description",
        help="Description of the knowledge base (e.g. 'A knowledge base for ice cream')",
    )
    attributes = st.text_input(
        "Attributes",
        help="Attributes of the knowledge base (e.g. 'flavor, viscosity, color')",
    )
    main_class = st.text_input(
        "Main Class",
        help="Main class of the knowledge base, without spaces (e.g. 'IceCreamType')",
    )
    generate_data = st.checkbox(
        "Generate data",
        help="""
        If checked, after generating the schema, generate example data.
        """,
    )
    if st.button("Make Schema"):
        st.write(f"Generating schema for *{kb_name}*")
        config_dict = {
            "kb_name": kb_name,
            "description": description,
            "attributes": attributes,
            "main_class": main_class,
        }
        config = KnowledgeBaseSpecification(**config_dict)
        ao = bootstrap_agent.bootstrap_schema(config)
        schema_dict = ao.model_dump()
        st.write(schema_dict)

        if generate_data:
            st.write(f"Generating data for *{kb_name}*")
            data = bootstrap_agent.bootstrap_data(schema=schema_dict)
            st.code(data, language="yaml")

elif option == CART:
    page_state = state.get_page_state(CART)
    st.subheader("Your items")
    for i, item in enumerate(cart.items):
        st.write(f"## Item {i}")
        item_obj = item.object
        st.code(yaml.dump(item_obj, sort_keys=False), language="yaml")
        if st.button(f"Remove {i} from cart"):
            cart.remove(item)

elif option == EXAMPLES:
    page_state = state.get_page_state(EXAMPLES)
    cc = get_case_collection()
    st.subheader("Examples")
    st.code(yaml.dump(cc, sort_keys=False), language="yaml")


elif option == ABOUT:
    page_state = state.get_page_state(ABOUT)
    st.subheader("About this instance")
    st.write(
        f"**DB:** {type(db).__name__} schema: {db.schema_proxy.name if db.schema_proxy else None}"
    )
    st.write("Collections:")
    rows = []
    for cn in db.collections():
        meta = db.collection_metadata(collection_name=cn, include_derived=True)
        rows.append(meta.dict())
    st.table(rows)

elif option == HELP:
    page_state = state.get_page_state(HELP)
    st.subheader("About")
    st.write(
        "CurateGPT is a tool for generating new entries for a knowledge base, assisted by LLMs."
    )
    st.write(
        "It is a highly generic system, but it's likely the instance"
        "you are using now is configured to work with ontologies."
    )
    st.subheader("Issues")
    st.write(
        "If you have any issues, please raise them on the"
        "[GitHub issue tracker](https://github.com/monarch-initiative/curategpt)."
    )
    st.subheader("Warning!")
    st.caption("CurateGPT is pre-alpha, documentation is incomplete!")
    st.caption(
        "If you are using a publicly deployed instance, some operations may be slow, or broken"
    )
    st.subheader("Instructions")
    st.write("Use the sidebar to select the operation you want to perform.")
    st.write(" * Synthesize: the core operation. Generate a new entry for the selected collection.")
    st.write(" * Chat: chat to a structured knowledge base or unstructured source.")
    st.write(" * Search: Search the vector stores.")
    st.write(
        " * Insert: Manually add data (do this responsibly if on public instance - no auth yet!."
    )
    st.write(" * About: View metadata for each instance.")
    st.subheader("FAQ")
    st.write("### Why are there no IDs?")
    st.write("LLMs will hallucinate IDs so we transform to CamelCase labels for demo purposes.")
    st.write("In future versions we will have a more elegant solution.")
    st.write("### What is the PubMed collection?")
    st.write("This is a special *virtual* collections. It is not populated ahead of time.")
    st.write("When this is used as a source, the pubmed API is called with a relevancy search.")
    st.write("These results are then combined with others to answer the query.")
    st.write("### What is the 'background' collection?")
    st.write(f"This is used only by '{EXTRACT}' to provide additional context.")



"""EUtils-based wrapper for studies in NCBI."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, List

import yaml

from curategpt.wrappers.literature.eutils_wrapper import EUtilsWrapper

logger = logging.getLogger(__name__)


@dataclass
class PMCWrapper(EUtilsWrapper):
    """
    A wrapper to provide a search facade over PMC via eutils.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "pmc"

    eutils_db: ClassVar[str] = "pmc"

    id_prefix: ClassVar[str] = "pmc"

    default_object_type = "Reference"

    fetch_tool = "esummary"

    def objects_from_dict(self, results: Dict) -> List[Dict]:
        print(yaml.dump(results))
        raise NotImplementedError()

        objs = []
        # TODO
        # _articles = results["pmc-articleset"]["article"]
        return objs



"""Chat with a KB."""

import logging
from dataclasses import dataclass
from typing import ClassVar, Dict, Iterable, Iterator, Optional

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


@dataclass
class OBOFormatWrapper(BaseWrapper):
    """
    A wrapper to index ontologies in OBO Format.

    Note that in contrast to the OntologyWrapper, this preserves OBO syntax stanzas, e.g.

    ```json
    {"term": "[Term]\nid: ...\nname: ...\n..."}
    ```
    """

    name: ClassVar[str] = "oboformat"

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        """
        Yield all objects in the view.

        :return:
        """
        path = self.source_locator

        with open(path) as file:
            obj = None
            for line in file.readlines():
                if line.startswith("["):
                    yield from self.wrap_object(obj)
                    obj = ""
                if obj is not None:
                    obj += line
            yield from self.wrap_object(obj)

    def wrap_object(self, obj: Optional[str]) -> Dict:
        if obj:
            yield {"term": obj}



"""Retrieval Augmented Generation (RAG) Base Class."""

import logging
from abc import ABC, abstractmethod
from copy import copy
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

import llm
from linkml_runtime import SchemaView
from pydantic import BaseModel as BaseModel
from pydantic import ConfigDict

from curategpt.store.schema_proxy import SchemaProxy

logger = logging.getLogger(__name__)


class AnnotatedObject(BaseModel):
    """
    Annotated object shadows a basic dictionary object
    """

    model_config = ConfigDict(protected_namespaces=())

    # object: Union[Dict[str, Any], List[Dict[str, Any]]] = {} - TODO: pydantic bug?
    object: Any = {}
    annotations: Dict[str, Any] = {}
    key_values: Dict[str, "AnnotatedObject"] = {}

    def as_single_object(self) -> Dict[str, Any]:
        """
        Return as a standard dictionary object.

        Each annotation is prefixed with an underscore.

        :return: dictionary object
        """
        object = copy(self.object)
        for key, value in self.annotations.items():
            object[f"_{key}"] = value
        for key, value in self.key_values.items():
            object[key] = value.as_single_object()
        return object

    @property
    def text(self) -> Optional[str]:
        """
        Get the text annotation of the object.

        :return:
        """
        return self.annotations.get("text", None)


EXAMPLE = Tuple[str, AnnotatedObject]


@dataclass
class Extractor(ABC):
    # db_adapter: DBAdapter = None
    schema_proxy: SchemaProxy = None
    model_name: str = None
    _model = None
    api_key: str = None
    raise_error_if_unparsable: bool = False

    @abstractmethod
    def extract(
        self, text: str, target_class: str, examples: List[AnnotatedObject] = None, **kwargs
    ) -> AnnotatedObject:
        """
        Schema-guided extraction

        :param text:
        :param kwargs:
        :return:
        """

    @property
    def model(self):
        """
        Get the model

        :param model_name:
        :return:
        """
        if self._model is None:
            model = llm.get_model(self.model_name or "gpt-3.5-turbo")
            if model.needs_key:
                model.key = llm.get_key(self.api_key, model.needs_key, model.key_env_var)
            self._model = model
        return self._model

    @property
    def schemaview(self) -> SchemaView:
        return self.schema_proxy.schemaview

    @property
    def pydantic_root_model(self) -> BaseModel:
        return self.schema_proxy.pydantic_root_model

    def deserialize(self, text: str, **kwargs) -> AnnotatedObject:
        """
        Deserialize text into an annotated object

        :param text:
        :return:
        """
        raise NotImplementedError



"""Evaluation utilities."""

from copy import copy
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict


class Outcome(BaseModel):

    model_config = ConfigDict(protected_namespaces=())

    prediction: Union[Dict[str, Any], List[Dict[str, Any]]] = {}
    expected: Union[Dict[str, Any], List[Dict[str, Any]]] = {}
    parameters: Dict[str, Any] = {}
    tp: int = 0
    tn: int = 0
    fp: int = 0
    fn: int = 0

    precision: Optional[float] = None
    recall: Optional[float] = None
    f1: Optional[float] = None

    by_field: Dict[str, int] = {}
    ixn_by_field: Dict[str, List[str]] = {}

    def calculate_metrics(self):
        self.precision = self.tp / (self.tp + self.fp) if self.tp + self.fp > 0 else 0
        self.recall = self.tp / (self.tp + self.fn) if self.tp + self.fn > 0 else 0
        self.f1 = (
            2 * self.precision * self.recall / (self.precision + self.recall)
            if self.precision + self.recall > 0
            else 0
        )

    def append_outcomes(self, outcomes: List["Outcome"]) -> None:
        for sub_outcome in outcomes:
            self.tp += sub_outcome.tp
            self.fp += sub_outcome.fp
            self.fn += sub_outcome.fn
            for key, value in sub_outcome.by_field.items():
                self.by_field[key] = self.by_field.get(key, 0) + value
            for key, value in sub_outcome.ixn_by_field.items():
                curr = set(self.ixn_by_field.get(key, []))
                self.ixn_by_field[key] = list(curr.union(value))
        self.calculate_metrics()

    def flatten(self) -> Dict[str, Any]:
        obj = self.model_dump()
        for k, v in copy(obj).items():
            if k == "parameters":
                obj.update(v)
                del obj[k]
            elif isinstance(v, dict):
                del obj[k]
            elif isinstance(v, list):
                obj[k] = [x for x in v if x]
        return obj


def score_prediction(
    predicted: Union[Dict, List], expected: Union[Dict, List], exclude: List = None
) -> Outcome:
    """
    Score the predicted activity.

    >>> outcome = score_prediction({"x": 1}, {"x": 1})
    >>> outcome.tp
    1

    >>> outcome = score_prediction([{"x": 1}], {"x": 1})
    >>> outcome.tp
    1


    >>> outcome = score_prediction({"x": 1}, {"x": 2})
    >>> outcome.tp
    0
    >>> outcome.recall
    0.0

    >>> outcome = score_prediction({"x": 1, "y": 2}, {"x": 1})
    >>> outcome.tp
    1
    >>> outcome.fp
    1

    >>> outcome = score_prediction([{"x": 1}, {"y": 1}], {"x": 1})
    >>> outcome.tp
    1
    >>> outcome.fp
    1


    :param predicted: The predicted activity
    :param expected: The expected activity
    :return: The score
    """
    if exclude is None:
        exclude = ["reference_title", "reference"]
    if isinstance(expected, list) or isinstance(predicted, list):
        if isinstance(expected, dict):
            expected = [expected]
        if isinstance(predicted, dict):
            predicted = [predicted]
        outcomes = best_matches(predicted, expected)
        outcome = Outcome(prediction=predicted, expected=expected)
        for sub_outcome in outcomes:
            outcome.tp += sub_outcome.tp
            outcome.fp += sub_outcome.fp
            outcome.fn += sub_outcome.fn
            for key, value in sub_outcome.by_field.items():
                outcome.by_field[key] = outcome.by_field.get(key, 0) + value
            for key, value in sub_outcome.ixn_by_field.items():
                outcome.ixn_by_field[key] = list(
                    set(outcome.ixn_by_field.get(key, [])).union(value)
                )
        outcome.calculate_metrics()
        return outcome
    outcome = Outcome(prediction=predicted, expected=expected)
    all_keys = set(predicted.keys()).union(expected.keys()).difference(exclude)
    for key in all_keys:
        if key in predicted and key in expected:
            if key == "relationships":
                pred_rels = predicted[key]
                exp_rels = expected[key]
                sub_outcomes = best_matches(pred_rels, exp_rels)
                n_tps = 0
                ixn = set()
                for sub_outcome in sub_outcomes:
                    outcome.tp += sub_outcome.tp
                    outcome.fp += sub_outcome.fp
                    outcome.fn += sub_outcome.fn
                    n_tps += sub_outcome.tp
                    if sub_outcome.precision == 1.0:
                        ixn = ixn.union({str(predicted[key])})
                outcome.by_field[key] = outcome.by_field.get(key, 0) + n_tps
                outcome.ixn_by_field[key] = list(set(outcome.ixn_by_field.get(key, [])).union(ixn))
                continue
            if predicted[key] == expected[key]:
                outcome.tp += 1
                outcome.by_field[key] = outcome.by_field.get(key, 0) + 1
                outcome.ixn_by_field[key] = list(
                    set(outcome.ixn_by_field.get(key, [])).union({predicted[key]})
                )
            else:
                outcome.fp += 1
                outcome.fn += 1
        elif key in predicted:
            outcome.fp += 1
        else:
            outcome.fn += 1
    outcome.calculate_metrics()
    return outcome


def best_matches(pred_rels, exp_rels) -> List[Outcome]:
    """
    Find the best matching pairs of relationships.

    Example:

    >>> outcomes = best_matches([], [])
    >>> len(outcomes)
    1
    >>> outcome = outcomes[0]
    >>> (outcome.tp, outcome.fp, outcome.fn)
    (0, 0, 0)
    >>> best_matches([{"x:": 1}], [])[0].precision
    0.0
    >>> outcome = best_matches([{"x": 1}], [{"x": 1}])[0]
    >>> outcome.precision
    1.0
    >>> outcome = best_matches([{"x": 1}], [{"y": 1}])[0]
    >>> outcome.precision
    0.0
    >>> pred_rels = [{"x":1}, {"y": 2}, {"z": 3}]
    >>> exp_rels = [{"y":2}, {"x": 1}, {"z": 3}]
    >>> outcomes = best_matches(pred_rels, exp_rels)
    >>> [o.precision for o in outcomes]
    [1.0, 1.0, 1.0]
    >>> exp_rels.append({"z": 4})
    >>> outcomes = best_matches(pred_rels, exp_rels)
    >>> sorted([o.precision for o in outcomes])
    [0.0, 1.0, 1.0, 1.0]

    """
    import numpy as np

    if not pred_rels:
        pred_rels = [{}]
    if not exp_rels:
        exp_rels = [{}]

    # Create a matrix to store the scores
    outcome_matrix = np.zeros((len(pred_rels), len(exp_rels)), dtype=object)
    outcome_ix = {}

    # Calculate the scores for each pair of pred_rel and exp_rel
    for i, pred_rel in enumerate(pred_rels):
        for j, exp_rel in enumerate(exp_rels):
            sub_outcome = score_prediction(pred_rel, exp_rel)
            outcome_matrix[i, j] = sub_outcome.tp
            outcome_ix[(i, j)] = sub_outcome

    # Find the best matching pairs
    outcomes = []
    max_row_indices = np.argmax(outcome_matrix, axis=1)
    max_col_indices = np.argmax(outcome_matrix, axis=0)
    best = []
    for i, _pred_rel in enumerate(pred_rels):
        best_j = max_row_indices[i]
        best.append((i, best_j))
        outcomes.append(outcome_ix[(i, best_j)])
    for j, _exp_rel in enumerate(exp_rels):
        best_i = max_col_indices[j]
        if (best_i, j) not in best:
            best.append((best_i, j))
            outcomes.append(outcome_ix[(best_i, j)])
    return outcomes



"""Chat with a filesystem."""

import glob
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Dict, Iterable, Iterator, Optional

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


@dataclass
class FilesystemWrapper(BaseWrapper):
    """
    A wrapper over a filesystem.

    This is a static wrapper: it cannot be searched
    """

    name: ClassVar[str] = "filesystem"
    root_directory: str = None
    glob: str = None

    skip_unprocessable = True

    max_text_length = 3000
    text_overlap = 200

    search_limit_multiplier: ClassVar[int] = 1

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        path = self.root_directory or "."
        if self.glob:
            files = glob.glob(os.path.join(path, "**", self.glob), recursive=True)
        else:
            files = []
            for dirpath, _dirnames, filenames in os.walk(path):
                for filename in filenames:
                    files.append(os.path.join(dirpath, filename))
        import textract

        for file in set(files):
            try:
                # Extract text from the file
                ex = file.split(".")[-1]
                if ex in ["py", "md"]:
                    text_utf8 = open(file).read()
                else:
                    text = textract.process(file)
                    text_utf8 = text.decode("utf-8")
                path = Path(file)
                stat = path.lstat()
                obj = {
                    "id": file,
                    "name": path.name,
                    "text": text_utf8,
                    "parent": str(path.parent),
                    "size": stat.st_size,
                    "mtime": stat.st_mtime,
                }
                yield from self.split_objects([obj])
            except Exception as e:
                if self.skip_unprocessable:
                    logger.warning(f"Failed to extract text from {file}. Reason: {e}")
                else:
                    raise e



"""Chat with a KB."""

import logging
import time
from dataclasses import dataclass
from typing import ClassVar, Dict, List, Optional

import requests

from curategpt.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "https://files.jgi.doe.gov/search/"


@dataclass
class JGIWrapper(BaseWrapper):
    """
    A wrapper to provide a search facade over JGI Data Search.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.
    """

    name: ClassVar[str] = "jgi"

    def external_search(
        self, text: str, expand: bool = False, where: Optional[Dict] = None, **kwargs
    ) -> List:
        params = {
            "x": 20,
        }
        if expand:

            def qt(t: str):
                t = t.strip()
                if " " in t:
                    return f'"{t}"'
                return t

            logger.info(f"Expanding search term: {text} to create JGI query")
            model = self.extractor.model
            response = model.prompt(
                text,
                system="""
                Take the specified search text, and expand it to a list
                of key terms used to construct a query. You will return results as
                semi-colon separated list of the most relevant terms. Make sure to
                include all relevant concepts in the returned terms.""",
            )
            terms = response.text().split(";")
            logger.info(f"Expanded terms: {terms}")
            terms = [qt(t) for t in terms]
            # terms = terms[0:2]
            search_term = "|".join(terms)
        else:
            search_term = text
        params["q"] = search_term
        if where:
            params.update(where)
        time.sleep(0.5)
        logger.info(f"Constructed query: {params}")
        response = requests.get(BASE_URL, params=params)
        data = response.json()

        organisms = data["organisms"]
        logger.info(f"Got {len(organisms)} organisms")
        results = []
        fields = [
            "doi",
            "name",
            "label",
            "title",
            "product_search_category",
            "status",
        ]
        for org in organisms:
            result = {}
            for k in fields:
                if k in org:
                    result[k] = org[k]
            files = org.get("files", [])
            if files:
                top_file = files[0]
                md = top_file.get("metadata", {})
                gold = md.get("gold", {})
                if gold:
                    result["display_name"] = gold.get("display_name", None)
                    result["ecosystem"] = gold.get("ecosystem", None)
            results.append(result)
        return results



from pathlib import Path
from typing import Dict, List, Optional, TextIO, Tuple, Union

from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel

from curategpt.store.duckdb_result import DuckDBSearchResult

OBJECT = Union[YAMLRoot, BaseModel, Dict, DuckDBSearchResult]
QUERY = Union[str, YAMLRoot, BaseModel, Dict, DuckDBSearchResult]
PROJECTION = Union[str, List[str]]
DEFAULT_COLLECTION = "test_collection"
SEARCH_RESULT = Tuple[DuckDBSearchResult, Dict, float, Optional[Dict]]
FILE_LIKE = Union[str, TextIO, Path]

IDS = "ids"
METADATAS = "metadatas"
EMBEDDINGS = "embeddings"
DOCUMENTS = "documents"
DISTANCES = "distances"

MODEL_MAP = {
    "text-embedding-ada-002": ("ada-002", 1536),
    "text-embedding-3-small": ("3-small", 1536),
    "text-embedding-3-large": ("3-large", 3072),
    "text-embedding-3-small-512": ("3-small-512", 512),
    "text-embedding-3-large-256": ("3-large-256", 256),
    "text-embedding-3-large-1024": ("3-large-1024", 1024),
}

DEFAULT_OPENAI_MODEL = "text-embedding-ada-002"
DEFAULT_MODEL = {"all-MiniLM-L6-v2": 384}



import logging
import random
from typing import Any, Dict, List, Optional, Union

from curategpt import ChromaDBAdapter, DBAdapter
from curategpt.evaluation.evaluation_datamodel import StratifiedCollection

logger = logging.getLogger(__name__)


def stratify_collection(
    store: DBAdapter,
    collection: str,
    num_training: Optional[int] = None,
    num_testing: Optional[int] = None,
    num_validation=0,
    testing_identifiers: Optional[List[str]] = None,
    fields_to_predict: Optional[Union[str, List[str]]] = None,
    ratio=0.7,
    where: Optional[Dict[str, Any]] = None,
) -> StratifiedCollection:
    """
    Stratifies a collection into training, testing, and validation sets.

    :param store:
    :param collection:
    :param num_training:
    :param num_testing:
    :param num_validation:
    :param fields_to_predict:
    :param ratio:
    :param where:
    :return:
    """
    if where is None:
        where = {}
    cm = store.collection_metadata(collection, include_derived=True)
    size = cm.object_count
    if testing_identifiers is None:
        if num_training is None and num_testing is None:
            num_training = int(size * ratio)
        if not num_testing:
            num_testing = size - num_training
        if not num_training:
            raise ValueError("Must specify either num_training or num_test")
        if num_training + num_testing > size:
            raise ValueError(f"num_training + num_test must be less than size {size}")
    # objs = list(store.peek(collection, limit=num_training + num_testing + num_validation))
    logger.info(f"Stratifying collection {collection} where={where}")
    if where:
        objs = list(store.find(where=where, collection=collection, limit=1000000))
    else:
        # chromadb doesn't do well with large limits on open-ended queries
        objs = list(store.peek(collection, limit=1000000))
    if fields_to_predict:
        if isinstance(fields_to_predict, str):
            fields_to_predict = fields_to_predict.split(",")
        objs = [obj for obj in objs if all(f in obj for f in fields_to_predict)]
    # randomize order w shuffle
    random.shuffle(objs)
    if testing_identifiers:
        logger.info(f"Taking from {len(testing_identifiers)} identifiers for testing set")

        def _is_in_test_set(obj: dict) -> bool:
            for fld in [store.identifier_field(collection), "original_id"]:
                if fld in obj and obj[fld] in testing_identifiers:
                    return True
            return False

        testing_set = [obj for obj in objs if _is_in_test_set(obj)]
        if not testing_set:
            raise ValueError(f"No testing objects found with identifiers: {testing_identifiers}")
        training_set = [obj for obj in objs if not _is_in_test_set(obj)]
        if num_testing:
            if len(testing_set) < num_testing:
                raise ValueError(
                    f"Only {len(testing_set)} testing objects found with identifiers: {testing_identifiers}"
                )
            testing_set = testing_set[:num_testing]
            logging.info(f"Using {len(testing_set)} testing objects")
        validation_set = []
    else:
        training_set = objs[:num_training]
        testing_set = objs[num_training : num_training + num_testing]
        validation_set = objs[
            num_training + num_testing : num_training + num_testing + num_validation
        ]
    logger.info(f"Training set size: {len(training_set)}")
    logger.info(f"Testing set size: {len(testing_set)}")
    logger.info(f"Validation set size: {len(validation_set)}")
    sc = StratifiedCollection(
        source=collection,
        training_set=training_set,
        testing_set=testing_set,
        validation_set=validation_set,
    )
    return sc


def stratify_collection_to_store(
    store: DBAdapter, collection: str, output_path: str, embedding_model=None, force=False, **kwargs
) -> Dict[str, str]:
    """
    Stratifies a collection into training, testing, and validation sets.

    Each collection is persisted to a separate collection in the output_path.

    :param store:
    :param collection:
    :param output_path:
    :param embedding_model:
    :param force:
    :param kwargs:
    :return:
    """
    sc = stratify_collection(store, collection=collection, **kwargs)
    output_db = ChromaDBAdapter(output_path)
    collections = {}
    existing_collections = output_db.list_collection_names()
    for sn in ["training", "testing", "validation"]:
        objs = getattr(sc, f"{sn}_set", [])
        size = len(objs)
        cn = f"{collection}_{sn}_{size}"
        collections[sn] = cn
        logger.info(f"Writing {size} objects to {cn}")
        if cn in existing_collections:
            logger.info(f"Collection {cn} already exists")
            if not force:
                logger.info(f"Reusing existing collection {cn}")
                continue
            else:
                logger.info(f"Overwriting existing collection {cn}")
        output_db.remove_collection(cn, exists_ok=True)
        output_db.insert(objs, collection=cn, model=embedding_model)
    return collections



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict


class CartItem(BaseModel):
    """
    A cart item is a single item in a cart
    """

    model_config = ConfigDict(protected_namespaces=())
    object: Union[Dict, BaseModel]
    object_type: Optional[str] = None
    source: Optional[str] = None
    metadata: Optional[Dict] = None


class Cart(BaseModel):
    """
    A cart is a list of items that can be added to or removed from
    """

    model_config = ConfigDict(protected_namespaces=())
    items: List[CartItem] = []

    @property
    def size(self):
        return len(self.items)

    def add(self, item: Union[CartItem, BaseModel, Dict]):
        if isinstance(item, dict):
            item = CartItem(object=item)
        if not isinstance(item, CartItem):
            if isinstance(item, BaseModel):
                item = CartItem(object=item.dict())
            elif isinstance(item, str):
                item = CartItem(object={"text": item})
            else:
                raise ValueError(f"Invalid cart item: {item}")
        self.items.append(item)

    def remove(self, item: CartItem):
        self.items = [i for i in self.items if i != item]



"""Chat with a KB."""

import logging
import tarfile
import tempfile
import time
from dataclasses import dataclass, field
from typing import ClassVar, Dict, List, Optional
from urllib.parse import urlparse
from urllib.request import urlretrieve

import requests
import requests_cache
from bs4 import BeautifulSoup
from defusedxml.ElementTree import fromstring
from eutils import Client

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)

ESEARCH_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
EFETCH_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"

RATE_LIMIT_DELAY = 1.0


def extract_all_text(element):
    text = element.text or ""
    for subelement in element:
        text += extract_all_text(subelement)
        text += subelement.tail if subelement.tail else ""
    return text


def extract_text_from_xml(xml_content):
    root = fromstring(xml_content)
    return extract_all_text(root).strip()


# TODO: rewrite to subclass EUtilsWrapper
@dataclass
class PubmedWrapper(BaseWrapper):
    """
    A wrapper to provide a search facade over PubMed.

    This is a dynamic wrapper: it can be used as a search facade,
    but cannot be ingested in whole.

    This can be used to restrict search to a journal; first find the journal:
    https://www.ncbi.nlm.nih.gov/nlmcatalog/?term=%22Case+Reports+in+Genetics%22

    E.g.

    `NLM Title Abbreviation: Case Rep Genet`

    Add this to the where clause:

    {"ta": "Case Rep Genet"}

    On the command line:

    .. code-block:: bash

        curategpt view search -V pubmed -I "where: {ta: Case Rep Genet}" "leukodystrophy and ischemia"
    """

    name: ClassVar[str] = "pubmed"

    eutils_client: Client = None

    session: requests.Session = field(default_factory=lambda: requests.Session())

    where: Optional[Dict] = None

    email: Optional[str] = None

    ncbi_key: Optional[str] = None

    is_fetch_full_text: Optional[bool] = None

    _uses_cache: bool = None

    def set_cache(self, name: str) -> None:
        self.session = requests_cache.CachedSession(name)
        self._uses_cache = True

    def external_search(
        self, text: str, expand: bool = True, where: Optional[Dict] = None, **kwargs
    ) -> List:
        if expand:
            logger.info(f"Expanding search term: {text} to create pubmed query")
            model = self.extractor.model
            response = model.prompt(
                text,
                system="""
                Take the specified search text, and expand it to a list
                of terms used to construct a PubMed query. You will return results as
                semi-colon separated list of the most relevant terms. Make sure to
                include all relevant concepts in the returned terms.""",
            )
            terms = response.text().split(";")
            search_term = " OR ".join(terms)
        else:
            search_term = text
        if not where and self.where:
            where = self.where
        if where:
            whereq = " AND ".join([f"{v}[{k}]" for k, v in where.items()])
            search_term = f"({search_term}) AND {whereq}"
        logger.info(f"Constructed search term: {search_term}")
        # Parameters for the request
        params = {
            "db": "pubmed",
            "term": search_term,
            "retmax": 100,
            "sort": "relevance",
            "retmode": "json",
        }

        # Note: we don't cache this call as there could be many
        # different search terms
        response = requests.get(ESEARCH_URL, params=params)
        time.sleep(RATE_LIMIT_DELAY)
        if response.status_code != 200:
            logger.error(f"Failed to search for {text}; params: {params}")
            return []
        data = response.json()

        # Extract PubMed IDs from the response
        if "esearchresult" not in data:
            logger.error(f"Failed to find results for {text}")
            logger.error(f"Data: {data}")
            return []
        pubmed_ids = data["esearchresult"]["idlist"]

        if not pubmed_ids:
            logger.warning(f"No results with {search_term}")
            if expand:
                logger.info("Trying again without expansion")
                return self.external_search(text, expand=False, **kwargs)
            else:
                logger.error(f"Failed to find results for {text}")
                return []

        logger.info(f"Found {len(pubmed_ids)} results: {pubmed_ids}")
        return self.objects_by_ids(pubmed_ids)

    def objects_by_ids(self, object_ids: List[str]) -> List[Dict]:
        pubmed_ids = sorted([x.replace("PMID:", "") for x in object_ids])
        session = self.session
        logger.debug(f"Using session: {session} [cached: {self._uses_cache} for {pubmed_ids}")

        # Parameters for the efetch request
        efetch_params = {
            "db": "pubmed",
            "id": ",".join(pubmed_ids),  # Combine PubMed IDs into a comma-separated string
            "rettype": "medline",
            "retmode": "text",
        }
        efetch_response = session.get(EFETCH_URL, params=efetch_params)
        if not self._uses_cache or not efetch_response.from_cache:
            # throttle if not using cache or if not cached
            logger.debug(f"Sleeping for {RATE_LIMIT_DELAY} seconds")
            time.sleep(RATE_LIMIT_DELAY)
        if not efetch_response.ok:
            logger.error(f"Failed to fetch data for {pubmed_ids}")
            raise ValueError(
                f"Failed to fetch data for {pubmed_ids} using {session} and {efetch_params}"
            )
        medline_records = efetch_response.text

        # Parsing titles and abstracts from the MEDLINE records
        parsed_data = []
        current_record = {}
        current_field = None

        for line in medline_records.split("\n"):
            if line.startswith("PMID- "):
                current_field = "id"
                current_record[current_field] = "PMID:" + line.replace("PMID- ", "").strip()
            if line.startswith("PMC - "):
                current_field = "pmcid"
                pmcid = line.replace("PMC - ", "").strip()
                current_record[current_field] = "PMCID:" + pmcid
                if self.is_fetch_full_text:
                    full_text = self.pmc_full_text(pmcid)
                    # full_text = self.fetch_full_text(pmcid)
                    current_record["full_text"] = full_text
            elif line.startswith("TI  - "):
                current_field = "title"
                current_record[current_field] = line.replace("TI  - ", "").strip()
            elif line.startswith("AB  - "):
                current_field = "abstract"
                current_record[current_field] = line.replace("AB  - ", "").strip()
            elif line.startswith("    "):  # Continuation of the previous field
                if current_field and current_field in current_record:
                    current_record[current_field] += " " + line.strip()
            else:
                current_field = None

            if line == "":
                if current_record:
                    parsed_data.append(current_record)
                    current_record = {}
        return parsed_data

    def fetch_pmcid(self, pmid: str) -> Optional[str]:
        pmid = pmid.replace("PMID:", "")
        session = self.session
        params = {"db": "pmc", "linkname": "pubmed_pmc", "id": pmid}
        url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi"
        response = session.get(url, params=params)
        root = fromstring(response.content)

        pmcid = None
        for link_set in root.findall(".//LinkSet"):
            for link in link_set.findall(".//Link"):
                pmcid = link.find("./Id").text
                if pmcid:
                    return f"PMC:PMC{pmcid}"
        return None

    def fetch_full_text(self, object_id: str) -> Optional[str]:
        # OLD
        session = self.session
        if object_id.startswith("PMID:"):
            pmcid = self.fetch_pmcid(object_id)
        else:
            pmcid = object_id

        if not pmcid:
            return None

        # PMC is a banana - get rid of the PMC prefix as well as local prefix
        pmcid = pmcid.replace("PMC:", "")
        pmcid = pmcid.replace("PMC", "")
        url = f"https://www.ncbi.nlm.nih.gov/pmc/utils/oa/oa.fcgi?id=PMC{pmcid}"
        response = session.get(url)
        root = fromstring(response.content)

        for record in root.findall(".//record"):
            for link in record.findall(".//link"):
                format_type = link.attrib.get("format")
                download_url = link.attrib.get("href")
                if format_type == "xml":
                    xml_response = requests.get(download_url)
                    return xml_response.text
                elif format_type == "tgz":
                    # make a named temp file
                    local_file_path = tempfile.NamedTemporaryFile().name
                    parsed_url = urlparse(download_url)
                    if parsed_url.scheme not in ["http", "https", "ftp"]:
                        continue
                    urlretrieve(download_url, local_file_path)  # noqa S310

                    # Open and extract the tar.gz file
                    with tarfile.open(local_file_path, "r:gz") as tar:
                        for member in tar.getmembers():
                            if member.name.endswith(".xml") or member.name.endswith(".nxml"):
                                f = tar.extractfile(member)
                                xml_str = f.read().decode("utf-8")
                                return extract_text_from_xml(xml_str)
                                # return xml_str
        return None

    def pmc_full_text(self, object_id: str) -> Optional[str]:
        if object_id.startswith("PMID:"):
            pmcid = self.fetch_pmcid(object_id)
        else:
            pmcid = object_id
        fulltext = self.pmc_xml(pmcid)
        logger.debug(f"FULL TEXT XML: {fulltext}")
        fullsoup = BeautifulSoup(fulltext, "xml")
        if fullsoup.find("pmc-articleset").find("article").find("body"):
            body = fullsoup.find("pmc-articleset").find("article").find("body").text
            return body
        else:
            return None

    def pmc_xml(self, pmc_id: str) -> str:
        """Get the text of one PubMed Central entry.

        Don't parse further here - just get the raw response.
        :param pmc_id: List of PubMed IDs, or string with single PMID
        :return: the text of a single entry as XML
        """
        session = self.session

        pmc_id = pmc_id.replace("PMC:", "")

        params = {
            "db": "pmc",
            "id": pmc_id,
            "rettype": "xml",
            "retmode": "xml",
        }

        if self.email and self.ncbi_key:
            params["api_key"] = self.ncbi_key
            params["email"] = self.email

        efetch_response = session.get(EFETCH_URL, params=params)
        if not self._uses_cache or not efetch_response.from_cache:
            # throttle if not using cache or if not cached
            logger.debug(f"Sleeping for {RATE_LIMIT_DELAY} seconds")
            time.sleep(RATE_LIMIT_DELAY)
        if not efetch_response.ok:
            logger.error(f"Failed to fetch data for {pmc_id}")
            raise ValueError(f"Failed to fetch data for {pmc_id} using {session} and {params}")
        return efetch_response.text



"""Abstract DB adapter."""

import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Dict, Iterable, Iterator, List, Optional, TextIO, Union

import pandas as pd
import yaml
from click.utils import LazyFile
from jsonlines import jsonlines

from curategpt.store.metadata import CollectionMetadata
from curategpt.store.schema_proxy import SchemaProxy
from curategpt.store.vocab import (
    DEFAULT_COLLECTION,
    DOCUMENTS,
    EMBEDDINGS,
    FILE_LIKE,
    METADATAS,
    OBJECT,
    PROJECTION,
    QUERY,
    SEARCH_RESULT,
)

logger = logging.getLogger(__name__)


def _get_file(file: Optional[FILE_LIKE] = None, mode="r") -> Optional[TextIO]:
    if file is None:
        return None
    if isinstance(file, Path):
        file = str(file)
    if isinstance(file, str):
        return open(file, mode)
    elif isinstance(file, TextIO):
        return file
    elif isinstance(file, LazyFile):
        return file
    else:
        raise TypeError(f"Unknown file type: {type(file)}")


@dataclass
class DBAdapter(ABC):
    """
    Base class for stores.

    This base class provides a common interface for a wide variety of document or object stores.
    The interface is intended to closely mimic the kind of interface found for document stores
    such as mongoDB or vector databases such as ChromaDB, but the intention is that can be
    used for SQL databases, SPARQL endpoints, or even file systems.

    The store allows for storage and retrieval of *objects* which are arbitrary dictionary
    objects, equivalient to a JSON object.

    Objects are partitioned into *collections*, which maps to the equivalent concept in
    MongoDB and ChromaDB.

    >>> from curategpt.store import get_store
    >>> store = get_store("in_memory")
    >>> store.insert({"name": "John", "age": 42}, collection="people")

    If you are used to working with MongoDB and ChromaDB APIs directly, one difference is that
    here we do not provide a separate Collection object, everything is handled through the
    store object. You can optionally bind a store object to a collection, which effectively
    gives you a collection object:

    >>> from curategpt.store import get_store
    >>> store = get_store("in_memory")
    >>> store.set_collection("people")
    >>> store.insert({"name": "John", "age": 42})

    TODO: decide if this is the final interface
    """

    path: str = None
    """Path to a location where the database is stored or disk or the network."""

    name: ClassVar[str] = "base"

    schema_proxy: Optional[SchemaProxy] = None
    """Schema manager"""

    collection: Optional[str] = None
    """Default collection"""

    # _field_names_by_collection: Dict[str, Optional[List[str]]] = field(default_factory=dict)
    _field_names_by_collection: Dict[str, Optional[List[str]]] = None

    # CUD operations

    @abstractmethod
    def insert(self, objs: Union[OBJECT, Iterable[OBJECT]], collection: str = None, **kwargs):
        """
        Insert an object or list of objects into the store.

        >>> from curategpt.store import get_store
        >>> store = get_store("in_memory")
        >>> store.insert([{"name": "John", "age": 42}], collection="people")

        :param objs:
        :param collection:
        :return:
        """

    def update(self, objs: Union[OBJECT, List[OBJECT]], collection: str = None, **kwargs):
        """
        Update an object or list of objects in the store.

        :param objs:
        :param collection:
        :return:
        """
        raise NotImplementedError

    def upsert(self, objs: Union[OBJECT, List[OBJECT]], collection: str = None, **kwargs):
        """
        Upsert an object or list of objects in the store.

        :param objs:
        :param collection:
        :return:
        """
        raise NotImplementedError

    def delete(self, id: str, collection: str = None, **kwargs):
        """
        Delete an object by its ID.

        :param id:
        :param collection:
        :return:
        """
        raise NotImplementedError

    # View operations

    def create_view(self, view_name: str, collection: str, expression: QUERY, **kwargs):
        """
        Create a view in the database.

        Todo:
        ----
        :param view:
        :return:
        """
        raise NotImplementedError

    # Collection operations

    def set_collection(self, collection: str):
        """
        Set the current collection.

        If this is set, then all subsequent operations will be performed on this collection, unless
        overridden.

        This allows the following

        >>> from curategpt.store import get_store
        >>> store = get_store("in_memory")
        >>> store.set_collection("people")
        >>> store.insert([{"name": "John", "age": 42}])

        to be written in place of

        >>> from curategpt.store import get_store
        >>> store = get_store("in_memory")
        >>> store.insert([{"name": "John", "age": 42}], collection="people")

        :param collection:
        :return:
        """
        self.collection = collection

    def _get_collection(self, collection: Optional[str] = None):
        if collection is not None:
            return collection
        elif self.collection is not None:
            return self.collection
        else:
            return DEFAULT_COLLECTION

    def remove_collection(self, collection: str = None, exists_ok=False, **kwargs):
        """
        Remove a collection from the database.

        :param collection:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def list_collection_names(self) -> List[str]:
        """
        List all collections in the database.

        :return: names of collections
        """

    @abstractmethod
    def collection_metadata(
        self, collection_name: Optional[str] = None, include_derived=False, **kwargs
    ) -> Optional[CollectionMetadata]:
        """
        Get the metadata for a collection.

        :param collection_name:
        :param include_derived: Include derived metadata, e.g. counts
        :return:
        """

    def set_collection_metadata(
        self, collection_name: Optional[str], metadata: CollectionMetadata, **kwargs
    ):
        """
        Set the metadata for a collection.

        >>> from curategpt.store import get_store
        >>> from curategpt.store import CollectionMetadata
        >>> store = get_store("in_memory")
        >>> cm = CollectionMetadata(name="People", description="People in the database")
        >>> store.set_collection_metadata("people", cm)

        :param collection_name:
        :return:
        """
        raise NotImplementedError

    def update_collection_metadata(self, collection_name: str, **kwargs) -> CollectionMetadata:
        """
        Update the metadata for a collection.

        :param collection_name:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def search(
        self, text: str, where: QUERY = None, collection: str = None, **kwargs
    ) -> Iterator[SEARCH_RESULT]:
        """
        Query the database for a text string.

        >>> from curategpt.store import get_store
        >>> store = get_store("chromadb", "db")
        >>> for obj, distance, info in store.search("forebrain neurons", collection="ont_cl"):
        ...     obj_id = obj["id"]
        ...     # print at precision of 2 decimal places
        ...     print(f"{obj_id} {distance:.2f}")
        <BLANKLINE>
        ...
        NeuronOfTheForebrain 0.28
        ...

        :param text:
        :param collection:
        :param where:
        :param kwargs:
        :return: tuple of object, distance, metadata
        """

    def find(
        self,
        where: QUERY = None,
        projection: PROJECTION = None,
        collection: str = None,
        **kwargs,
    ) -> Iterator[SEARCH_RESULT]:
        """
        Query the database.

        >>> from curategpt.store import get_store
        >>> store = get_store("chromadb", "db")
        >>> objs = list(store.find({"name": "NeuronOfTheForebrain"}, collection="ont_cl"))

        :param collection:
        :param where:
        :param projection:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def matches(self, obj: OBJECT, **kwargs) -> Iterator[SEARCH_RESULT]:
        """
        Query the database for matches to an object.

        :param obj:
        :param kwargs:
        :return:
        """

    @abstractmethod
    def lookup(self, id: str, collection: str = None, **kwargs) -> OBJECT:
        """
        Lookup an object by its ID.

        :param id:
        :param collection:
        :return:
        """

    def lookup_multiple(self, ids: List[str], **kwargs) -> Iterator[OBJECT]:
        """
        Lookup an object by its ID.

        :param id:
        :param collection:
        :return:
        """
        yield from [self.lookup(id, **kwargs) for id in ids]

    @abstractmethod
    def peek(self, collection: str = None, limit=5, **kwargs) -> Iterator[OBJECT]:
        """
        Peek at first N objects in a collection.

        :param collection:
        :param limit:
        :return:
        """
        raise NotImplementedError

    # Schema operations

    @abstractmethod
    def fetch_all_objects_memory_safe(
        self, collection: str = None, batch_size: int = 100, **kwargs
    ) -> Iterator[OBJECT]:
        """
        Fetch all objects from a collection, in batches to avoid memory overload.
        """
        raise NotImplementedError

    def identifier_field(self, collection: str = None) -> str:
        # TODO: use collection
        if self.schema_proxy and self.schema_proxy.schemaview:
            fields = []
            for s in self.schema_proxy.schemaview.all_slots(attributes=True).values():
                if s.identifier:
                    fields.append(s.name)
            if fields:
                if len(fields) > 1:
                    raise ValueError(f"Multiple identifier fields: {fields}")
                return fields[0]
        return "id"

    def label_field(self, collection: str = None) -> str:
        field_names = self.field_names(collection)
        preferred = ["label", "name", "title"]
        for candidate in preferred:
            if candidate in field_names:
                return candidate
        return field_names[0] if field_names else "label"

    # TODO: Might be abstract as currently gives not same values for DuckDBVSSAdapter and ChromaDBAdapter
    def field_names(self, collection: str = None) -> List[str]:
        """
        Return the names of all top level fields in the database for a collection.

        :param collection:
        :return:
        """
        # TODO: use schema proxy if set
        if not self._field_names_by_collection:
            self._field_names_by_collection = {}
        collection = self._get_collection(collection)
        if collection not in self._field_names_by_collection:
            logger.debug(f"Getting field names for {collection}")
            objs = self.peek(collection=collection, limit=1)
            if not objs:
                raise ValueError(f"Collection {collection} is empty")
            fields = []
            for obj in objs:
                for f in obj:
                    if f not in fields:
                        fields.append(f)
            self._field_names_by_collection[collection] = fields
        else:
            logger.debug(f"Using cached field names for {collection}")
        return self._field_names_by_collection[collection]

    # Loading and dumping
    def dump(
        self,
        collection: str = None,
        to_file: FILE_LIKE = None,
        metadata_to_file: FILE_LIKE = None,
        format=None,
        include=None,
        **kwargs,
    ):
        """
        Dump the database to a file.

        :param collection:
        :param kwargs:
        :return:
        """
        collection_name = collection
        collection = self._get_collection(collection)
        logger.info(f"Dumping collection {self.collection_metadata(collection) }")
        metadata = self.collection_metadata(collection).model_dump(exclude_none=True)
        if format is None:
            format = "json"
        if not include:
            include = [EMBEDDINGS, DOCUMENTS, METADATAS]
            # include = ["embeddings", "documents", "metadatas"]
        if not isinstance(include, list):
            include = list(include)
        objects = list(self.find(collection=collection, include=include, **kwargs))
        if format.startswith("venomx"):
            import venomx as vx
            from venomx.tools.file_io import save_index

            cm = self.collection_metadata(collection_name, include_derived=False)
            label_field = self.label_field(collection_name)
            id_field = self.identifier_field(collection_name)

            def _label(obj):
                return obj.get(label_field, None)

            def _id(obj):
                original_id = obj.get("original_id", None)
                if original_id:
                    return original_id
                return obj.get(id_field, None)

            rows = [
                {"id": _id(obj), "text": meta["document"], "values": meta["_embeddings"]}
                for obj, _, meta in objects
            ]
            vx_objects = [{"id": _id(obj), "label": _label(obj)} for obj, _, _ in objects]
            logger.info(f"Num vx objects: {len(vx_objects)}")
            vx_index = vx.Index(
                embedding_model=vx.model.venomx.Model(name=cm.model),
                dataset=vx.model.venomx.Dataset(name=collection_name),
                embeddings_frame=pd.DataFrame(rows),
                objects=vx_objects,
            )
            save_index(vx_index, to_file)
            return
        to_file = _get_file(to_file, "w")
        metadata_to_file = _get_file(metadata_to_file, "w")
        streaming = True
        if format == "jsonl":
            writer = jsonlines.Writer(to_file)
            writer.write_all(objects)
            writer.close()
        elif format == "yamlblock":
            for obj in objects:
                to_file.write("---\n")
                yaml.dump(obj, to_file)
        else:
            streaming = False
            database = {"metadata": metadata, "objects": list(objects)}
            if format == "json":
                json.dump(database, to_file)
            elif format == "yaml":
                yaml.dump(database, to_file)
            else:
                raise ValueError(f"Unknown format {format}")
        if streaming:
            if not metadata_to_file:
                raise ValueError("Streaming dump requires metadata_to_file")
            if format == "jsonl":
                metadata_to_file.write(json.dumps(metadata))
            elif format == "yamlblock":
                metadata_to_file.write(yaml.dump(metadata))

    def dump_then_load(self, collection: str = None, target: "DBAdapter" = None):
        """
        Dump a collection to a file, then load it into another database.

        :param collection:
        :param target:
        :return:
        """
        raise NotImplementedError



"""Chat with a KB."""

import logging
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Iterable, Iterator, List, Optional

import requests_cache
from oaklib import BasicOntologyInterface

from curategpt.wrappers import BaseWrapper

logger = logging.getLogger(__name__)

BASE_URL = "https://reactome.org/ContentService/data"


def ids_from_tree(objs: List):
    """
    Recursively yield all ids from a tree of objects

    Note: may contain duplicates

    :param objs:
    :return:
    """
    for obj in objs:
        yield "Reactome:" + obj["stId"]
        if "children" in obj:
            yield from ids_from_tree(obj["children"])


def term_object(obj: dict):
    return {
        "id": obj["databaseName"] + ":" + obj["accession"],
        "label": obj["name"],
    }


def pub_object(obj: dict):
    if "pubMedIdentifier" not in obj:
        logger.warning(f"Missing pubmed id for {obj}")
        return None
    return {
        "id": "PMID:" + str(obj["pubMedIdentifier"]),
        "title": obj["title"],
    }


def simple_entity_object(obj: dict):
    return {
        "id": "Reactome:" + str(obj["stId"]),
        "label": obj["displayName"],
        "type": obj.get("referenceType", None),
    }


def generic_object(obj: dict):
    return obj["displayName"]


OBJECT_FUNCTION_MAP = {
    "compartment": term_object,
    "literatureReference": pub_object,
    "catalystActivity": generic_object,
    "input": simple_entity_object,
    "output": simple_entity_object,
    "hasComponent": simple_entity_object,
    "precedingEvent": simple_entity_object,
    "goBiologicalProcess": term_object,
}


@dataclass
class ReactomeWrapper(BaseWrapper):
    """
    A wrapper over a Reactome API.
    """

    name: ClassVar[str] = "reactome"

    prefix = "Reactome"

    _label_adapter: BasicOntologyInterface = None

    default_object_type = "Event"

    taxon_id: str = field(default="NCBITaxon:9606")

    def object_ids(self, taxon_id: str = None, **kwargs) -> Iterator[str]:
        """
        Get all object ids for a given taxon id

        :param taxon_id:
        :param kwargs:
        :return:
        """
        session = requests_cache.CachedSession("reactome")

        if not taxon_id:
            taxon_id = self.taxon_id
        taxon_id_acc = taxon_id.split(":")[1]
        response = session.get(f"{BASE_URL}/eventsHierarchy/{taxon_id_acc}/")
        response.raise_for_status()
        obj = response.json()
        yield from ids_from_tree(obj)

    def objects(
        self, collection: str = None, object_ids: Optional[Iterable[str]] = None, **kwargs
    ) -> Iterator[Dict]:
        """
        All events

        :param collection:
        :param object_ids:
        :param kwargs:
        :return:
        """
        session = requests_cache.CachedSession("reactome")

        if not object_ids:
            object_ids = self.object_ids(**kwargs)
        else:
            object_ids = object_ids

        visited = set()
        for object_id in object_ids:
            if object_id in visited:
                continue
            visited.add(object_id)
            logger.info(f"Getting {object_id}")
            local_id = object_id.split(":")[1]
            response = session.get(f"{BASE_URL}/query/{local_id}")
            obj = response.json()
            summations = obj["summation"]
            new_obj = {
                "id": object_id,
                "label": obj["displayName"],
                "speciesName": obj["speciesName"],
                "description": "\n".join([x["text"] for x in summations]),
                "type": obj["schemaClass"],
            }
            for key, func in OBJECT_FUNCTION_MAP.items():
                if key in obj:
                    new_obj[key] = [func(x) for x in obj[key] if isinstance(x, dict)]
            yield new_obj



import logging
from typing import Iterator, Tuple

from curategpt import DBAdapter
from curategpt.utils.vector_algorithms import compute_cosine_similarity, top_matches

logger = logging.getLogger(__name__)


def match_collections(
    db: DBAdapter, left_collection: str, right_collection: str, other_db: DBAdapter = None
) -> Iterator[Tuple[dict, dict, float]]:
    """
    Match every element in left collection with every element in right collection.

    Currently this returns best matches for left collection only

    :param db:
    :param left_collection:
    :param right_collection:
    :param other_db: optional - defaults to main
    :return: tuple of object pair plus cosine similarity score
    """
    if not other_db:
        other_db = db
    include = ["metadatas", "documents", "embeddings"]
    logger.info(f"Querying left objects from {left_collection}")
    left_objs = list(db.find({}, collection=left_collection, include=include))
    logger.info(f"Querying right objects from {right_collection}")
    right_objs = list(other_db.find({}, collection=right_collection, include=include))
    left_vectors = [info["_embeddings"] for _, __, info in left_objs]
    right_vectors = [info["_embeddings"] for _, __, info in right_objs]
    logger.info(f"Computing cosine similarity for {len(left_vectors)} x {len(right_vectors)}")
    sim_matrix = compute_cosine_similarity(left_vectors, right_vectors)
    logger.info("Finding top matches")
    tm_ix, tm_vals = top_matches(sim_matrix)
    logger.info(f"Yielding {len(tm_ix)} matches")
    i = 0
    for ix, val in zip(tm_ix, tm_vals, strict=True):
        yield left_objs[i][0], right_objs[ix][0], val
        i += 1
